# Copyright 2008-2020 pydicom authors. See LICENSE file for details.
"""Hold DicomFile class, which does basic I/O for a dicom file."""

from io import BytesIO
import os
from struct import Struct
from types import TracebackType
from typing import TYPE_CHECKING, cast, Any, TypeVar, Protocol

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Callable


ExitException = tuple[
    type[BaseException] | None, BaseException | None, TracebackType | None
]
Self = TypeVar("Self", bound="DicomIO")


class ReadableBuffer(Protocol):
    def read(self, size: int = ..., /) -> bytes: ...  # pragma: no cover

    def seek(self, offset: int, whence: int = ..., /) -> int: ...  # pragma: no cover

    def tell(self) -> int: ...  # pragma: no cover


class WriteableBuffer(Protocol):
    def seek(self, offset: int, whence: int = ..., /) -> int: ...  # pragma: no cover

    def tell(self) -> int: ...  # pragma: no cover

    def write(
        self, b: bytes | bytearray | memoryview, /
    ) -> int: ...  # pragma: no cover


class DicomIO:
    """Wrapper for managing buffer-like objects used when reading or writing
    DICOM datasets.
    """

    def __init__(self, buffer: ReadableBuffer | WriteableBuffer) -> None:
        """Create a new ``DicomIO`` instance.

        Parameters
        ----------
        buffer : buffer-like object
            A buffer-like object that implements:

            * ``seek()`` and ``tell()`` methods with the same signatures as
              :meth:`io.IOBase.seek` and :meth:`io.IOBase.tell`
            * a ``read()`` method with the same signature as
              :meth:`io.RawIOBase.read` if it supports reading data from
              itself, and/or
            * a ``write()`` method with the same signature as
              :meth:`io.RawIOBase.write` if it supports writing data to itself

            If `buffer` supports reading it can be used with
            :func:`~pydicom.filereader.dcmread` as the source to decode a DICOM
            dataset from, and if it supports writing it can be used with
            :func:`~pydicom.filewriter.dcmwrite` as the destination for the
            encoded DICOM dataset.
        """
        # Data packers/unpackers
        self._us_unpacker: Callable[[bytes], tuple[Any, ...]]
        self._us_packer: Callable[[int], bytes]
        self._ul_unpacker: Callable[[bytes], tuple[Any, ...]]
        self._ul_packer: Callable[[int], bytes]
        self._tag_unpacker: Callable[[bytes], tuple[Any, ...]]
        self._tag_packer: Callable[[int, int], bytes]

        # Store the encoding method
        self._implicit_vr: bool
        self._little_endian: bool

        # The buffer-like object being wrapped
        self._buffer = buffer

        # The filename associated with the buffer-like
        self._name: str | None = getattr(self._buffer, "name", None)

        # It's more efficient to replace the existing class methods
        #   instead of wrapping them
        if hasattr(buffer, "read"):
            self.read = buffer.read

        if hasattr(buffer, "write"):
            self.write = buffer.write

        if hasattr(buffer, "close"):
            self.close = buffer.close

        # seek() and tell() are always required
        self.seek = buffer.seek
        self.tell = buffer.tell

    def close(self, *args: Any, **kwargs: Any) -> Any:
        """Close the buffer (if possible)"""
        pass

    def __enter__(self: Self) -> Self:
        return self

    def __exit__(self, *exc_info: ExitException) -> None:
        self.close()

    @property
    def is_little_endian(self) -> bool:
        """Get/set the endianness for encoding/decoding, ``True`` for little
        endian and ``False`` for big endian.
        """
        if not hasattr(self, "_little_endian"):
            raise AttributeError(
                f"{type(self).__name__}.is_little_endian' has not been set"
            )

        return self._little_endian

    @is_little_endian.setter
    def is_little_endian(self, value: bool) -> None:
        if not isinstance(value, bool):
            raise TypeError(f"'{type(self).__name__}.is_little_endian' must be bool")

        self._little_endian = value

        endianness = "><"[value]
        self._us_packer = Struct(f"{endianness}H").pack
        self._us_unpacker = Struct(f"{endianness}H").unpack
        self._ul_packer = Struct(f"{endianness}L").pack
        self._ul_unpacker = Struct(f"{endianness}L").unpack
        self._tag_packer = Struct(f"{endianness}2H").pack
        self._tag_unpacker = Struct(f"{endianness}2H").unpack

    @property
    def is_implicit_VR(self) -> bool:
        """Get/set the VR mode for encoding/decoding. ``True`` for implicit VR
        and ``False`` for explicit VR.
        """
        if not hasattr(self, "_implicit_vr"):
            raise AttributeError(
                f"{type(self).__name__}.is_implicit_VR' has not been set"
            )

        return self._implicit_vr

    @is_implicit_VR.setter
    def is_implicit_VR(self, value: bool) -> None:
        if not isinstance(value, bool):
            raise TypeError(f"'{type(self).__name__}.is_implicit_VR' must be bool")

        self._implicit_vr = value

    @property
    def name(self) -> str | None:
        """Return the value of the :attr:`~pydicom.filebase.DicomIO.parent`'s
        ``name`` attribute, or ``None`` if no such attribute.
        """
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        self._name = name

    @property
    def parent(self) -> ReadableBuffer | WriteableBuffer:
        """Return the buffer object being wrapped."""
        return self._buffer

    def read(self, size: int = -1, /) -> bytes:
        """Read up to `size` bytes from the buffer and return them. If `size`
        is unspecified, all bytes until EOF are returned.

        Fewer than `size` bytes may be returned if the operating system call
        returns fewer than `size` bytes.
        """
        raise TypeError(
            f"'{type(self).__name__}' cannot be used with "
            f"'{type(self._buffer).__name__}': object has no read() method"
        )

    def read_exact(self, length: int, nr_retries: int = 3) -> bytes:
        """Return `length` bytes read from the buffer.

        Parameters
        ----------
        length : int
            The number of bytes to be read. If ``None`` (default) then read all
            the bytes available.
        nr_retries : int, optional
            The number of tries to read data when the number of bytes read
            from the buffer is less than `length`. Default ``3``.

        Returns
        -------
        bytes
            The read data.

        Raises
        ------
        EOFError
            If unable to read `length` bytes.
        """
        bytes_read = self.read(length)
        if len(bytes_read) == length:
            return bytes_read

        # Use a bytearray because concatenating bytes is expensive
        bytes_read = bytearray(bytes_read)
        attempts = 0
        while (num_bytes := len(bytes_read)) < length and attempts < nr_retries:
            bytes_read += self.read(length - num_bytes)
            attempts += 1

        if num_bytes == length:
            return bytes(bytes_read)

        raise EOFError(
            f"Unexpected end of file. Read {num_bytes} bytes of {length} "
            f"expected starting at position 0x{self.tell() - num_bytes:x}"
        )

    def read_tag(self) -> tuple[int, int]:
        """Return a DICOM tag value read from the buffer."""
        return cast(
            tuple[int, int],
            self._tag_unpacker(self.read_exact(4)),
        )

    def read_UL(self) -> int:
        """Return a UL value read from the buffer."""
        return cast(int, self._ul_unpacker(self.read(4))[0])

    def read_US(self) -> int:
        """Return a US value read from the buffer."""
        return cast(int, self._us_unpacker(self.read(2))[0])

    def seek(self, offset: int, whence: int = os.SEEK_SET, /) -> int:
        """Change the buffer position to the given byte `offset`, relative to
        the position indicated by `whence` and return the new absolute position.
        """
        raise NotImplementedError()  # pragma: no cover

    def tell(self) -> int:
        """Return the current stream position of the buffer"""
        raise NotImplementedError()  # pragma: no cover

    def write(self, b: bytes | bytearray | memoryview, /) -> int:
        """Write the bytes-like object `b` to the buffer and return the number
        of bytes written.
        """
        raise TypeError(
            f"'{type(self).__name__}' cannot be used with "
            f"'{type(self._buffer).__name__}': object has no write() method"
        )

    def write_tag(self, tag: int) -> None:
        """Write a DICOM tag to the buffer."""
        self.write(self._tag_packer(tag >> 16, tag & 0xFFFF))

    def write_UL(self, val: int) -> None:
        """Write a UL value to the buffer."""
        self.write(self._ul_packer(val))

    def write_US(self, val: int) -> None:
        """Write a US value to the buffer."""
        self.write(self._us_packer(val))


class DicomFileLike(DicomIO):
    """Wrapper for file-likes to simplify encoding/decoding DICOM datasets.

    See Also
    --------
    :class:`~pydicom.filebase.DicomIO`
    :class:`~pydicom.filebase.DicomBytesIO`
    """

    pass


def DicomFile(*args: Any, **kwargs: Any) -> DicomFileLike:
    """Return an opened :class:`~pydicom.filebase.DicomFileLike` from a file-like."""
    return DicomFileLike(open(*args, **kwargs))


class DicomBytesIO(DicomIO):
    """Wrapper for :class:`io.BytesIO` to simplify encoding/decoding DICOM datasets.

    See Also
    --------
    :class:`~pydicom.filebase.DicomIO`
    :class:`~pydicom.filebase.DicomFileLike`
    """

    def __init__(self, initial_bytes: bytes | bytearray | memoryview = b"") -> None:
        """Create a new DicomBytesIO instance.

        Parameters
        ----------
        buffer : bytes | bytearray | memoryview, optional
            The buffer to write to or read from, default is an empty buffer.
        """
        buffer = BytesIO(initial_bytes)
        super().__init__(buffer)

        self.getvalue = buffer.getvalue



# Copyright 2008-2024 pydicom authors. See LICENSE file for details.

from typing import Any, TYPE_CHECKING, Protocol
from collections.abc import MutableSequence, Callable

from pydicom import config
from pydicom.datadict import dictionary_VR, private_dictionary_VR
from pydicom.errors import BytesLengthException
from pydicom.misc import warn_and_log
from pydicom.tag import BaseTag, _LUT_DESCRIPTOR_TAGS
from pydicom.valuerep import VR


if TYPE_CHECKING:  # pragma: no cover
    from pydicom.dataset import Dataset
    from pydicom.dataelem import RawDataElement

    class RawDataHook(Protocol):
        def __call__(
            self,
            raw: RawDataElement,
            data: dict[str, Any],
            *,
            encoding: str | MutableSequence[str] | None = ...,
            ds: Dataset | None = ...,
            **kwargs: Any,
        ) -> None: ...


class Hooks:
    """Management class for callback functions.

    .. versionadded:: 3.0

    .. warning::

        New instances of this class should not be created, instead use the instance
        imported with ``from pydicom.hooks import hooks``.

    **Available Hooks**

    For conversion of raw element data to :class:`~pydicom.dataelem.DataElement`
    using :func:`~pydicom.dataelem.convert_raw_data_element`:

    * ``"raw_element_vr"``: function to perform VR lookups for raw elements,
      default :func:`raw_element_vr`.
    * ``"raw_element_value"``: function to convert raw element values to types
      appropriate for the VR, default :func:`raw_element_value`.
    * ``"raw_element_kwargs"``: `kwargs` :class:`dict` passed to the callback
      functions.
    """

    def __init__(self) -> None:
        """Initialize a new ``Hooks`` instance."""
        self.raw_element_value: RawDataHook
        self.raw_element_vr: RawDataHook
        self.raw_element_kwargs: dict[str, Any] = {}

    def register_callback(self, hook: str, func: Callable) -> None:
        """Register the callback function `func` to a hook.

        Example
        -------

        .. code-block:: python

            from pydicom import dcmread
            from pydicom.hooks import hooks, raw_element_value_fix_separator

            hooks.register_callback(
                "raw_element_value", raw_element_value_fix_separator
            )
            kwargs = {"target_VRs": ("DS", "IS")}
            hooks.register_kwargs("raw_element_kwargs", kwargs)

            ds = dcmread("path/to/dataset.dcm")

        Parameters
        ----------
        hook : str
            The name of the hook to register the function to, allowed values
            ``"raw_element_vr"`` and ``"raw_element_value"``.
        func : Callable
            The callback function to use with the hook. Only one callback function can
            be used per hook. For details on the required function signatures please
            see the documentation for the corresponding calling function.
        """
        if not callable(func):
            raise TypeError("'func' must be a callable function")

        if hook == "raw_element_value":
            self.raw_element_value = func
        elif hook == "raw_element_vr":
            self.raw_element_vr = func
        else:
            raise ValueError(f"Unknown hook '{hook}'")

    def register_kwargs(self, hook: str, kwargs: dict[str, Any]) -> None:
        """Register a `kwargs` :class:`dict` to be passed to the corresponding
        callback function(s).

        Parameters
        ----------
        hook : str
            The name of the hook to register `kwargs` to, allowed value
            ``"raw_element_kwargs"``.
        kwargs : dict[str, Any]
            A :class:`dict` containing keyword arguments to be passed to the
            hook's corresponding callback function(s).
        """
        if not isinstance(kwargs, dict):
            raise TypeError(f"'kwargs' must be a dict, not '{type(kwargs).__name__}'")

        if hook == "raw_element_kwargs":
            self.raw_element_kwargs = kwargs
        else:
            raise ValueError(f"Unknown hook '{hook}'")


def _private_vr_for_tag(ds: "Dataset | None", tag: BaseTag) -> str:
    """Return the VR for a known private tag, otherwise "UN".

    Parameters
    ----------
    ds : Dataset, optional
        The dataset needed for the private creator lookup.
        If not given, "UN" is returned.
    tag : BaseTag
        The private tag to lookup. The caller has to ensure that the
        tag is private.

    Returns
    -------
    str
        "LO" if the tag is a private creator, the VR of the private tag if
        found in the private dictionary, or "UN".
    """
    if tag.is_private_creator:
        return VR.LO

    # invalid private tags are handled as UN
    if ds is not None and (tag.element & 0xFF00):
        private_creator_tag = tag.group << 16 | (tag.element >> 8)
        private_creator = ds.get(private_creator_tag, "")
        if private_creator:
            try:
                return private_dictionary_VR(tag, private_creator.value)
            except KeyError:
                pass

    return VR.UN


def raw_element_vr(
    raw: "RawDataElement",
    data: dict[str, Any],
    *,
    encoding: str | MutableSequence[str] | None = None,
    ds: "Dataset | None" = None,
    **kwargs: Any,
) -> None:
    """Determine the VR to use for `raw`.

    .. versionadded:: 3.0

    Default callback function for the ``"raw_element_vr"`` hook.

    Parameters
    ----------
    raw : RawDataElement
        The raw data element to determine the VR for.
    data : dict[str, Any]
        A dict to store the results of the VR lookup, which should be added
        as ``{"VR": str}``.
    ds : pydicom.dataset.Dataset | None
        The dataset the element belongs to.
    **kwargs: dict[str, Any]
        Additional keyword arguments.
    """
    vr = raw.VR
    if vr is None:  # Can be if was implicit VR
        try:
            vr = dictionary_VR(raw.tag)
        except KeyError:
            # just read the bytes, no way to know what they mean
            if raw.tag.is_private:
                # for VR for private tags see PS3.5, 6.2.2
                vr = _private_vr_for_tag(ds, raw.tag)

            # group length tag implied in versions < 3.0
            elif raw.tag.element == 0:
                vr = VR.UL
            else:
                msg = f"VR lookup failed for the raw element with tag {raw.tag}"
                if config.settings.reading_validation_mode == config.RAISE:
                    raise KeyError(msg)

                vr = VR.UN
                warn_and_log(f"{msg} - setting VR to 'UN'")
    elif vr == VR.UN and config.replace_un_with_known_vr:
        # handle rare case of incorrectly set 'UN' in explicit encoding
        # see also DataElement.__init__()
        if raw.tag.is_private:
            vr = _private_vr_for_tag(ds, raw.tag)
        elif raw.value is None or len(raw.value) < 0xFFFF:
            try:
                vr = dictionary_VR(raw.tag)
            except KeyError:
                pass

    data["VR"] = vr


def raw_element_value(
    raw: "RawDataElement",
    data: dict[str, Any],
    *,
    encoding: str | MutableSequence[str] | None = None,
    ds: "Dataset | None" = None,
    **kwargs: Any,
) -> None:
    """Convert the encoded value for `raw` to an appropriate type.

    .. versionadded:: 3.0

    Will set the VR to **UN** if unable to convert the value.

    Default callback function for the ``"raw_element_value"`` hook.

    Parameters
    ----------
    raw : RawDataElement
        The raw data element to determine the value for.
    data : dict[str, Any]
        A dict to store the results of the value conversion, which should be added
        as ``{"value": Any}``.
    encoding : str | MutableSequence[str] | None
        The character set encoding to use for text VRs.
    **kwargs: dict[str, Any]
        Additional keyword arguments.
    """
    from pydicom.values import convert_value

    vr = data["VR"]
    try:
        value = convert_value(vr, raw, encoding)
    except NotImplementedError as exc:
        raise NotImplementedError(f"{exc} in tag {raw.tag}")
    except BytesLengthException as exc:
        # Failed conversion, either raise or convert to a UN VR
        msg = (
            f"{exc} This occurred while trying to parse {raw.tag} according "
            f"to VR '{raw.VR}'."
        )
        if not config.convert_wrong_length_to_UN:
            raise BytesLengthException(
                f"{msg} To replace this error with a warning set "
                "pydicom.config.convert_wrong_length_to_UN = True."
            )

        warn_and_log(f"{msg} Setting VR to 'UN'.")
        data["VR"] = VR.UN
        value = raw.value

    if raw.tag in _LUT_DESCRIPTOR_TAGS:
        # We only fix the first value as the third value is 8 or 16
        if value and isinstance(value, list):
            try:
                if value[0] < 0:
                    value[0] += 65536
            except Exception:
                pass

    data["value"] = value


def raw_element_value_fix_separator(
    raw: "RawDataElement",
    data: dict[str, Any],
    *,
    encoding: str | MutableSequence[str] | None = None,
    ds: "Dataset | None",
    separator: str | bytes = b",",
    target_VRs: tuple[str, ...] | None = None,
    **kwargs: Any,
) -> None:
    """Convenience function to fix values with an invalid multivalue separator.

    .. versionadded:: 3.0

    Alternative callback function for the ``"raw_element_value"`` hook.

    Example
    -------

    Fix **DS** and **IS** elements that use an invalid ":" character as the
    multivalue separator::

        from pydicom import dcmread
        from pydicom.hooks import hooks, raw_element_value_fix_separator

        hooks.register_callback(
            "raw_element_value",
            raw_element_value_fix_separator,
        )
        hooks.register_kwargs(
            "raw_element",
            {"target_VRs": ("DS", "IS"), "separator": b":"},
        )

        ds = dcmread("path/to/dataset.dcm")


    Parameters
    ----------
    raw : RawDataElement
        The raw data element to determine the value for.
    data : dict[str, Any]
        A dict to store the results of the value conversion, which should be added
        as ``{"value": Any}``.
    encoding : str | MutableSequence[str] | None
        The character set encoding to use for text VRs.
    separator : str | bytes, optional
        The invalid separator character to be replaced by an ASCII backslash (0x5C).
    target_VRs : tuple[str, ...], optional
        The VRs the fix should apply.
    **kwargs: dict[str, Any]
        Additional keyword arguments.
    """
    vr = data["VR"]
    if target_VRs and vr in target_VRs and isinstance(raw.value, bytes):
        if isinstance(separator, str):
            separator = separator.encode("ascii")

        raw = raw._replace(value=raw.value.replace(separator, b"\x5C"))

    raw_element_value(raw, data, encoding=encoding, ds=ds, **kwargs)


def raw_element_value_retry(
    raw: "RawDataElement",
    data: dict[str, Any],
    *,
    encoding: str | MutableSequence[str] | None = None,
    ds: "Dataset | None",
    target_VRs: dict[str, tuple[str, ...]] | None = None,
    **kwargs: Any,
) -> None:
    """Convenience function to retry value conversion using a different VR.

    .. versionadded:: 3.0

    Alternative callback function for the ``"raw_element_value"`` hook.

    Example
    -------

    Retry the value conversion for **DS** elements using **US** or **SS**::

        from pydicom import dcmread
        from pydicom.hooks import hooks, raw_element_value_retry

        hooks.register_callback(
            "raw_element_value",
            raw_element_value_retry,
        )
        hooks.register_kwargs(
            "raw_element",
            {"target_VRs": {"DS": ("US", "SS")}},
        )

        ds = dcmread("path/to/dataset.dcm")

    Parameters
    ----------
    raw : RawDataElement
        The raw data element to determine the value for.
    data : dict[str, Any]
        A dict to store the results of the value conversion, which should be added
        as ``{"value": Any}``.
    encoding : str | MutableSequence[str] | None
        The character set encoding to use for text VRs.
    target_VRs : dict[str, tuple[str, ...]], optional
        The ``{VRs the fix should apply: tuple of alternative VRs to try}``.
    **kwargs: dict[str, Any]
        Additional keyword arguments.
    """
    from pydicom.values import convert_value

    try:
        raw_element_value(raw, data, encoding=encoding, ds=ds, **kwargs)
    except Exception as exc:
        vr = data["VR"]
        if target_VRs and vr in target_VRs:
            for candidate in target_VRs[vr]:
                try:
                    data["value"] = convert_value(candidate, raw)
                    data["VR"] = candidate
                    return
                except Exception:
                    pass

        raise exc


hooks: Hooks = Hooks()
"""The global :class:`~pydicom.hooks.Hooks` singleton.

.. versionadded:: 3.0
"""
hooks.raw_element_value = raw_element_value
hooks.raw_element_vr = raw_element_vr



# Copyright 2024 pydicom authors. See LICENSE file for details.
"""Use GDCM <https://github.com/malaterre/GDCM> to decompress encoded
*Pixel Data*.

This module is not intended to be used directly.
"""

from typing import cast

from pydicom import uid
from pydicom.pixels.decoders.base import DecodeRunner
from pydicom.pixels.common import PhotometricInterpretation as PI

try:
    import gdcm

    GDCM_VERSION = tuple(int(x) for x in gdcm.Version.GetVersion().split("."))
    HAVE_GDCM = True
except ImportError:
    HAVE_GDCM = False


DECODER_DEPENDENCIES = {
    uid.JPEGBaseline8Bit: ("gdcm>=3.0.10",),
    uid.JPEGExtended12Bit: ("gdcm>=3.0.10",),
    uid.JPEGLossless: ("gdcm>=3.0.10",),
    uid.JPEGLosslessSV1: ("gdcm>=3.0.10",),
    uid.JPEGLSLossless: ("gdcm>=3.0.10",),
    uid.JPEGLSNearLossless: ("gdcm>=3.0.10",),
    uid.JPEG2000Lossless: ("gdcm>=3.0.10",),
    uid.JPEG2000: ("gdcm>=3.0.10",),
}


def is_available(uid: str) -> bool:
    """Return ``True`` if a pixel data decoder for `uid` is available for use,
    ``False`` otherwise.
    """
    if not HAVE_GDCM or GDCM_VERSION < (3, 0):
        return False

    return uid in DECODER_DEPENDENCIES


def _decode_frame(src: bytes, runner: DecodeRunner) -> bytes:
    """Return the decoded `src` as :class:`bytes`.

    Parameters
    ----------
    src : bytes
        An encoded pixel data frame.
    runner : pydicom.pixels.decoders.base.DecodeRunner
        The runner managing the decoding.

    Returns
    -------
    bytes
        The decoded pixel data frame.
    """
    tsyntax = runner.transfer_syntax
    photometric_interpretation = runner.photometric_interpretation
    bits_stored = runner.bits_stored
    if tsyntax == uid.JPEGExtended12Bit and bits_stored != 8:
        raise NotImplementedError(
            "GDCM does not support 'JPEG Extended' for samples with 12-bit precision"
        )

    if (
        tsyntax == uid.JPEGLSNearLossless
        and runner.pixel_representation == 1
        and bits_stored < 8
    ):
        raise ValueError(
            "Unable to decode signed lossy JPEG-LS pixel data with a sample "
            "precision less than 8 bits"
        )

    if tsyntax in uid.JPEGLSTransferSyntaxes and bits_stored in (6, 7):
        raise ValueError(
            "Unable to decode unsigned JPEG-LS pixel data with a sample "
            "precision of 6 or 7 bits"
        )

    fragment = gdcm.Fragment()
    fragment.SetByteStringValue(src)

    fragments = gdcm.SequenceOfFragments.New()
    fragments.AddFragment(fragment)

    elem = gdcm.DataElement(gdcm.Tag(0x7FE0, 0x0010))
    elem.SetValue(fragments.__ref__())

    img = gdcm.Image()
    img.SetNumberOfDimensions(2)
    img.SetDimensions((runner.columns, runner.rows, 1))
    img.SetDataElement(elem)

    pi_type = gdcm.PhotometricInterpretation.GetPIType(photometric_interpretation)
    img.SetPhotometricInterpretation(gdcm.PhotometricInterpretation(pi_type))
    if runner.samples_per_pixel > 1:
        img.SetPlanarConfiguration(runner.planar_configuration)

    ts_type = gdcm.TransferSyntax.GetTSType(str.__str__(tsyntax))
    img.SetTransferSyntax(gdcm.TransferSyntax(ts_type))

    if tsyntax in uid.JPEGLSTransferSyntaxes:
        # GDCM always returns JPEG-LS data as color-by-pixel
        runner.set_option("planar_configuration", 0)
        bits_stored = runner.get_option("jls_precision", bits_stored)
        if 0 < bits_stored <= 8:
            runner.set_option("bits_allocated", 8)
        elif 8 < bits_stored <= 16:
            runner.set_option("bits_allocated", 16)

    if tsyntax in uid.JPEG2000TransferSyntaxes:
        # GDCM pixel container size is based on precision
        bits_stored = runner.get_option("j2k_precision", bits_stored)
        if 0 < bits_stored <= 8:
            runner.set_option("bits_allocated", 8)
        elif 8 < bits_stored <= 16:
            runner.set_option("bits_allocated", 16)
        elif 16 < bits_stored <= 32:
            runner.set_option("bits_allocated", 32)

    pixel_format = gdcm.PixelFormat(
        runner.samples_per_pixel,
        runner.bits_allocated,
        bits_stored,
        bits_stored - 1,
        runner.pixel_representation,
    )
    img.SetPixelFormat(pixel_format)

    # GDCM returns char* as str, so re-encode it to bytes
    frame = img.GetBuffer().encode("utf-8", "surrogateescape")

    # GDCM returns YBR_ICT and YBR_RCT as RGB
    if tsyntax in uid.JPEG2000TransferSyntaxes and photometric_interpretation in (
        PI.YBR_ICT,
        PI.YBR_RCT,
    ):
        runner.set_option("photometric_interpretation", PI.RGB)

    return cast(bytes, frame)



# Copyright 2008-2024 pydicom authors. See LICENSE file for details.
"""Use pyjpegls <https://github.com/pydicom/pyjpegls> to decompress encoded
*Pixel Data*.

This module is not intended to be used directly.
"""
from typing import cast

from pydicom import uid
from pydicom.pixels.utils import _passes_version_check
from pydicom.pixels.decoders.base import DecodeRunner

try:
    import jpeg_ls

except ImportError:
    pass


DECODER_DEPENDENCIES = {
    uid.JPEGLSLossless: ("numpy", "pyjpegls>=1.2"),
    uid.JPEGLSNearLossless: ("numpy", "pyjpegls>=1.2"),
}


def is_available(uid: str) -> bool:
    """Return ``True`` if the decoder has its dependencies met, ``False`` otherwise"""
    return _passes_version_check("jpeg_ls", (1, 2))


def _decode_frame(src: bytes, runner: DecodeRunner) -> bytearray:
    """Return the decoded image data in `src` as a :class:`bytearray`."""
    buffer, info = jpeg_ls.decode_pixel_data(src)
    # Interleave mode 0 is colour-by-plane, 1 and 2 are colour-by-pixel
    if info["components"] > 1:
        if info["interleave_mode"] == 0:
            runner.set_option("planar_configuration", 1)
        else:
            runner.set_option("planar_configuration", 0)

    precision = info["bits_per_sample"]
    if 0 < precision <= 8:
        runner.set_option("bits_allocated", 8)
    elif 8 < precision <= 16:
        runner.set_option("bits_allocated", 16)

    return cast(bytearray, buffer)



# Copyright 2008-2021 pydicom authors. See LICENSE file for details.
"""Functions related to writing DICOM data."""

from collections.abc import Sequence, MutableSequence, Iterable
from copy import deepcopy
from io import BufferedIOBase
from struct import pack
from typing import BinaryIO, Any, cast
from collections.abc import Callable
import zlib

from pydicom import config
from pydicom.charset import default_encoding, convert_encodings, encode_string
from pydicom.dataelem import (
    convert_raw_data_element,
    DataElement,
    RawDataElement,
)
from pydicom.dataset import Dataset, validate_file_meta, FileMetaDataset
from pydicom.filebase import DicomFile, DicomBytesIO, DicomIO, WriteableBuffer
from pydicom.fileutil import (
    path_from_pathlike,
    PathType,
    buffer_remaining,
    read_buffer,
    reset_buffer_position,
)
from pydicom.misc import warn_and_log
from pydicom.multival import MultiValue
from pydicom.tag import (
    Tag,
    BaseTag,
    ItemTag,
    ItemDelimiterTag,
    SequenceDelimiterTag,
    tag_in_exception,
    _LUT_DESCRIPTOR_TAGS,
)
from pydicom.uid import (
    DeflatedExplicitVRLittleEndian,
    UID,
    ImplicitVRLittleEndian,
    ExplicitVRBigEndian,
)
from pydicom.valuerep import (
    PersonName,
    IS,
    DSclass,
    DA,
    DT,
    TM,
    EXPLICIT_VR_LENGTH_32,
    VR,
    AMBIGUOUS_VR,
    CUSTOMIZABLE_CHARSET_VR,
)
from pydicom.values import convert_numbers

if config.have_numpy:
    import numpy

# Ambiguous VR Correction
# (0018,9810) Zero Velocity Pixel Value
# (0022,1452) Mapped Pixel Value
# (0028,0104)/(0028,0105) Smallest/Largest Valid Pixel Value
# (0028,0106)/(0028,0107) Smallest/Largest Image Pixel Value
# (0028,0108)/(0028,0109) Smallest/Largest Pixel Value in Series
# (0028,0110)/(0028,0111) Smallest/Largest Image Pixel Value in Plane
# (0028,0120) Pixel Padding Value
# (0028,0121) Pixel Padding Range Limit
# (0028,1101-1103) Red/Green/Blue Palette Color Lookup Table Descriptor
# (0028,3002) LUT Descriptor
# (0040,9216)/(0040,9211) Real World Value First/Last Value Mapped
# (0060,3004)/(0060,3006) Histogram First/Last Bin Value
_AMBIGUOUS_US_SS_TAGS = {
    0x00189810,
    0x00221452,
    0x00280104,
    0x00280105,
    0x00280106,
    0x00280107,
    0x00280108,
    0x00280109,
    0x00280110,
    0x00280111,
    0x00280120,
    0x00280121,
    0x00281101,
    0x00281102,
    0x00281103,
    0x00283002,
    0x00409211,
    0x00409216,
    0x00603004,
    0x00603006,
}

# (5400,0110) Channel Minimum Value
# (5400,0112) Channel Maximum Value
# (5400,100A) Waveform Padding Data
# (5400,1010) Waveform Data
_AMBIGUOUS_OB_OW_TAGS = {0x54000110, 0x54000112, 0x5400100A, 0x54001010}

# (60xx,3000) Overlay Data
_OVERLAY_DATA_TAGS = {x << 16 | 0x3000 for x in range(0x6000, 0x601F, 2)}


def _correct_ambiguous_vr_element(
    elem: DataElement,
    ancestors: list[Dataset],
    is_little_endian: bool,
) -> DataElement:
    """Implementation for `correct_ambiguous_vr_element`.
    See `correct_ambiguous_vr_element` for description.
    """
    # The zeroth dataset is the nearest, the last is the root dataset
    ds = ancestors[0]

    # 'OB or OW': 7fe0,0010 PixelData
    if elem.tag == 0x7FE00010:
        # Compressed Pixel Data
        # PS3.5 Annex A.4
        #   If encapsulated, VR is OB and length is undefined
        if elem.is_undefined_length:
            elem.VR = VR.OB
        elif ds.original_encoding[0]:
            # Non-compressed Pixel Data - Implicit Little Endian
            # PS3.5 Annex A1: VR is always OW
            elem.VR = VR.OW
        else:
            # Non-compressed Pixel Data - Explicit VR
            # PS3.5 Annex A.2:
            # If BitsAllocated is > 8 then VR shall be OW,
            # else may be OB or OW.
            # If we get here, the data has not been written before
            # or has been converted from Implicit Little Endian,
            # so we default to OB for BitsAllocated 1 or 8
            elem.VR = VR.OW if cast(int, ds.BitsAllocated) > 8 else VR.OB

    # 'US or SS' and dependent on PixelRepresentation
    elif elem.tag in _AMBIGUOUS_US_SS_TAGS:
        # US if PixelRepresentation value is 0x0000, else SS
        #   For references, see the list at
        #   https://github.com/pydicom/pydicom/pull/298
        # PixelRepresentation is usually set in the root dataset

        # If correcting during write, or after implicit read when the
        #   element is on the same level as pixel representation
        pixel_rep = next(
            (
                cast(int, x.PixelRepresentation)
                for x in ancestors
                if getattr(x, "PixelRepresentation", None) is not None
            ),
            None,
        )

        if pixel_rep is None:
            # If correcting after implicit read when the element isn't
            #   on the same level as pixel representation
            pixel_rep = next(
                (x._pixel_rep for x in ancestors if hasattr(x, "_pixel_rep")),
                None,
            )

        if pixel_rep is None:
            # If no pixel data is present, none if these tags is used,
            # so we can just ignore a missing PixelRepresentation in this case
            pixel_rep = 1
            if (
                "PixelRepresentation" not in ds
                and "PixelData" not in ds
                or ds.PixelRepresentation == 0
            ):
                pixel_rep = 0

        elem.VR = VR.US if pixel_rep == 0 else VR.SS
        byte_type = "H" if pixel_rep == 0 else "h"

        if elem.VM == 0:
            return elem

        # Need to handle type check for elements with VM > 1
        elem_value = elem.value if elem.VM == 1 else cast(Sequence[Any], elem.value)[0]
        if not isinstance(elem_value, int):
            elem.value = convert_numbers(
                cast(bytes, elem.value), is_little_endian, byte_type
            )

    # 'OB or OW' and dependent on WaveformBitsAllocated
    elif elem.tag in _AMBIGUOUS_OB_OW_TAGS:
        # If WaveformBitsAllocated is > 8 then OW, otherwise may be
        #   OB or OW.
        #   See PS3.3 C.10.9.1.
        if ds.original_encoding[0]:
            elem.VR = VR.OW
        else:
            elem.VR = VR.OW if cast(int, ds.WaveformBitsAllocated) > 8 else VR.OB

    # 'US or OW': 0028,3006 LUTData
    elif elem.tag == 0x00283006:
        # First value in LUT Descriptor is how many values in
        #   LUTData, if there's only one value then must be US
        # As per PS3.3 C.11.1.1.1
        if cast(Sequence[int], ds.LUTDescriptor)[0] == 1:
            elem.VR = VR.US
            if elem.VM == 0:
                return elem

            elem_value = (
                elem.value if elem.VM == 1 else cast(Sequence[Any], elem.value)[0]
            )
            if not isinstance(elem_value, int):
                elem.value = convert_numbers(
                    cast(bytes, elem.value), is_little_endian, "H"
                )
        else:
            elem.VR = VR.OW

    # 'OB or OW': 60xx,3000 OverlayData and dependent on Transfer Syntax
    elif elem.tag in _OVERLAY_DATA_TAGS:
        # Implicit VR must be OW, explicit VR may be OB or OW
        #   as per PS3.5 Section 8.1.2 and Annex A
        elem.VR = VR.OW

    return elem


def correct_ambiguous_vr_element(
    elem: DataElement | RawDataElement,
    ds: Dataset,
    is_little_endian: bool,
    ancestors: list[Dataset] | None = None,
) -> DataElement | RawDataElement:
    """Attempt to correct the ambiguous VR element `elem`.

    When it's not possible to correct the VR, the element will be returned
    unchanged. Currently the only ambiguous VR elements not corrected for are
    all retired or part of DICONDE.

    If the VR is corrected and is 'US' or 'SS' then the value will be updated
    using the :func:`~pydicom.values.convert_numbers` function.

    .. versionchanged:: 3.0

        The `ancestors` keyword argument was added.

    Parameters
    ----------
    elem : dataelem.DataElement or dataelem.RawDataElement
        The element with an ambiguous VR.
    ds : dataset.Dataset
        The dataset containing `elem`.
    is_little_endian : bool
        The byte ordering of the values in the dataset.
    ancestors : list[pydicom.dataset.Dataset] | None
        A list of the ancestor datasets to look through when trying to find
        the relevant element value to use in VR correction. Should be ordered
        from closest to furthest. If ``None`` then will build itself
        automatically starting at `ds` (default).

    Returns
    -------
    dataelem.DataElement or dataelem.RawDataElement
        The corrected element
    """
    ancestors = [ds] if ancestors is None else ancestors

    if elem.VR in AMBIGUOUS_VR:
        # convert raw data elements before handling them
        if isinstance(elem, RawDataElement):
            elem = convert_raw_data_element(elem, ds=ds)
            ds.__setitem__(elem.tag, elem)

        try:
            _correct_ambiguous_vr_element(elem, ancestors, is_little_endian)
        except AttributeError as e:
            raise AttributeError(
                f"Failed to resolve ambiguous VR for tag {elem.tag}: {e}"
            )

    return elem


def correct_ambiguous_vr(
    ds: Dataset,
    is_little_endian: bool,
    ancestors: list[Dataset] | None = None,
) -> Dataset:
    """Iterate through `ds` correcting ambiguous VR elements (if possible).

    When it's not possible to correct the VR, the element will be returned
    unchanged. Currently the only ambiguous VR elements not corrected for are
    all retired or part of DICONDE.

    If the VR is corrected and is 'US' or 'SS' then the value will be updated
    using the :func:`~pydicom.values.convert_numbers` function.

    .. versionchanged:: 3.0

        The `ancestors` keyword argument was added.

    Parameters
    ----------
    ds : pydicom.dataset.Dataset
        The dataset containing ambiguous VR elements.
    is_little_endian : bool
        The byte ordering of the values in the dataset.
    ancestors : list[pydicom.dataset.Dataset] | None
        A list of the ancestor datasets to look through when trying to find
        the relevant element value to use in VR correction. Should be ordered
        from closest to furthest. If ``None`` then will build itself
        automatically starting at `ds` (default).

    Returns
    -------
    ds : dataset.Dataset
        The corrected dataset

    Raises
    ------
    AttributeError
        If a tag is missing in `ds` that is required to resolve the ambiguity.
    """
    # Construct the tree if `ds` is the root level dataset
    # tree = Tree(ds) if tree is None else tree
    ancestors = [ds] if ancestors is None else ancestors

    # Iterate through the elements
    for elem in ds.elements():
        # raw data element sequences can be written as they are, because we
        # have ensured that the transfer syntax has not changed at this point
        if elem.VR == VR.SQ:
            elem = ds[elem.tag]
            for item in cast(MutableSequence["Dataset"], elem.value):
                ancestors.insert(0, item)
                correct_ambiguous_vr(item, is_little_endian, ancestors)
        elif elem.VR in AMBIGUOUS_VR:
            correct_ambiguous_vr_element(elem, ds, is_little_endian, ancestors)

    del ancestors[0]
    return ds


def write_numbers(fp: DicomIO, elem: DataElement, struct_format: str) -> None:
    """Write a "value" of type struct_format from the dicom file.

    "Value" can be more than one number.

    Parameters
    ----------
    fp : file-like
        The file-like to write the encoded data to.
    elem : dataelem.DataElement
        The element to encode.
    struct_format : str
        The character format as used by the struct module.
    """
    value = elem.value
    if value is None or value == "":
        return  # don't need to write anything for no or empty value

    endianChar = "><"[fp.is_little_endian]
    format_string = endianChar + struct_format
    try:
        try:
            # works only if list, not if string or number
            value.append
        except AttributeError:  # is a single value - the usual case
            fp.write(pack(format_string, value))
        else:
            # Some ambiguous VR elements ignore the VR for part of the value
            # e.g. LUT Descriptor is 'US or SS' and VM 3, but the first and
            #   third values are always US (the third should be <= 16, so SS is OK)
            if struct_format == "h" and elem.tag in _LUT_DESCRIPTOR_TAGS and value:
                fp.write(pack(f"{endianChar}H", value[0]))
                value = value[1:]

            fp.write(pack(f"{endianChar}{len(value)}{struct_format}", *value))
    except Exception as exc:
        raise OSError(f"{exc}\nfor data_element:\n{elem}")


def write_OBvalue(fp: DicomIO, elem: DataElement) -> None:
    """Write a data_element with VR of 'other byte' (OB)."""

    if elem.is_buffered:
        bytes_written = 0
        buffer = cast(BufferedIOBase, elem.value)
        with reset_buffer_position(buffer):
            for chunk in read_buffer(buffer):
                bytes_written += fp.write(chunk)
    else:
        bytes_written = fp.write(cast(bytes, elem.value))

    if bytes_written % 2:
        fp.write(b"\x00")


def write_OWvalue(fp: DicomIO, elem: DataElement) -> None:
    """Write a data_element with VR of 'other word' (OW).

    Note: This **does not currently do the byte swapping** for Endian state.
    """

    if elem.is_buffered:
        bytes_written = 0
        buffer = cast(BufferedIOBase, elem.value)
        with reset_buffer_position(buffer):
            for chunk in read_buffer(buffer):
                bytes_written += fp.write(chunk)
    else:
        bytes_written = fp.write(cast(bytes, elem.value))

    if bytes_written % 2:
        fp.write(b"\x00")


def write_UI(fp: DicomIO, elem: DataElement) -> None:
    """Write a data_element with VR of 'unique identifier' (UI)."""
    write_string(fp, elem, "\0")  # pad with 0-byte to even length


def _is_multi_value(val: Any) -> bool:
    """Return True if `val` is a multi-value container."""
    if config.have_numpy and isinstance(val, numpy.ndarray):
        return True

    return isinstance(val, MultiValue | list | tuple)


def multi_string(val: str | Iterable[str]) -> str:
    """Put a string together with delimiter if has more than one value"""
    if _is_multi_value(val):
        return "\\".join(val)

    return cast(str, val)


def write_PN(
    fp: DicomIO, elem: DataElement, encodings: list[str] | None = None
) -> None:
    if not encodings:
        encodings = [default_encoding]

    val: list[PersonName]
    if elem.VM == 1:
        val = [cast(PersonName, elem.value)]
    else:
        val = cast(list[PersonName], elem.value)

    enc = b"\\".join([elem.encode(encodings) for elem in val])
    if len(enc) % 2 != 0:
        enc += b" "

    fp.write(enc)


def write_string(fp: DicomIO, elem: DataElement, padding: str = " ") -> None:
    """Write a single or multivalued ASCII string."""
    val = multi_string(cast(str | Iterable[str], elem.value))
    if val is not None:
        if len(val) % 2 != 0:
            val += padding  # pad to even length

        if isinstance(val, str):
            val = val.encode(default_encoding)  # type: ignore[assignment]

        fp.write(val)  # type: ignore[arg-type]


def write_text(
    fp: DicomIO, elem: DataElement, encodings: list[str] | None = None
) -> None:
    """Write a single or multivalued text string."""
    encodings = encodings or [default_encoding]
    val = elem.value
    if val is not None:
        if _is_multi_value(val):
            val = cast(Sequence[bytes] | Sequence[str], val)
            if isinstance(val[0], str):
                val = cast(Sequence[str], val)
                val = b"\\".join([encode_string(val, encodings) for val in val])
            else:
                val = cast(Sequence[bytes], val)
                val = b"\\".join([val for val in val])
        else:
            val = cast(bytes | str, val)
            if isinstance(val, str):
                val = encode_string(val, encodings)

        if len(val) % 2 != 0:
            val = val + b" "  # pad to even length
        fp.write(val)


def write_number_string(fp: DicomIO, elem: DataElement) -> None:
    """Handle IS or DS VR - write a number stored as a string of digits."""
    # If the DS or IS has an original_string attribute, use that, so that
    # unchanged data elements are written with exact string as when read from
    # file
    val = elem.value
    if _is_multi_value(val):
        val = cast(Sequence[IS] | Sequence[DSclass], val)
        val = "\\".join(
            x.original_string if hasattr(x, "original_string") else str(x) for x in val
        )
    else:
        val = cast(IS | DSclass, val)
        if hasattr(val, "original_string"):
            val = val.original_string
        else:
            val = str(val)

    if len(val) % 2 != 0:
        val = val + " "  # pad to even length

    val = bytes(val, default_encoding)

    fp.write(val)


def _format_DA(val: DA | None) -> str:
    if val is None:
        return ""

    if hasattr(val, "original_string"):
        return val.original_string

    return val.strftime("%Y%m%d")


def write_DA(fp: DicomIO, elem: DataElement) -> None:
    val = elem.value
    if isinstance(val, str):
        write_string(fp, elem)
    else:
        if _is_multi_value(val):
            val = cast(Sequence[DA], val)
            val = "\\".join(x if isinstance(x, str) else _format_DA(x) for x in val)
        else:
            val = _format_DA(cast(DA, val))

        if len(val) % 2 != 0:
            val = val + " "  # pad to even length

        if isinstance(val, str):
            val = val.encode(default_encoding)

        fp.write(val)


def _format_DT(val: DT | None) -> str:
    if val is None:
        return ""

    if hasattr(val, "original_string"):
        return val.original_string

    if val.microsecond > 0:
        return val.strftime("%Y%m%d%H%M%S.%f%z")

    return val.strftime("%Y%m%d%H%M%S%z")


def write_DT(fp: DicomIO, elem: DataElement) -> None:
    val = elem.value
    if isinstance(val, str):
        write_string(fp, elem)
    else:
        if _is_multi_value(val):
            val = cast(Sequence[DT], val)
            val = "\\".join(x if isinstance(x, str) else _format_DT(x) for x in val)
        else:
            val = _format_DT(cast(DT, val))

        if len(val) % 2 != 0:
            val = val + " "  # pad to even length

        if isinstance(val, str):
            val = val.encode(default_encoding)

        fp.write(val)


def _format_TM(val: TM | None) -> str:
    if val is None:
        return ""

    if hasattr(val, "original_string"):
        return val.original_string

    if val.microsecond > 0:
        return val.strftime("%H%M%S.%f")

    return val.strftime("%H%M%S")


def write_TM(fp: DicomIO, elem: DataElement) -> None:
    val = elem.value
    if isinstance(val, str):
        write_string(fp, elem)
    else:
        if _is_multi_value(val):
            val = cast(Sequence[TM], val)
            val = "\\".join(x if isinstance(x, str) else _format_TM(x) for x in val)
        else:
            val = _format_TM(cast(TM, val))

        if len(val) % 2 != 0:
            val = val + " "  # pad to even length

        if isinstance(val, str):
            val = val.encode(default_encoding)

        fp.write(val)


def write_data_element(
    fp: DicomIO,
    elem: DataElement | RawDataElement,
    encodings: str | list[str] | None = None,
) -> None:
    """Write the data_element to file fp according to
    dicom media storage rules.
    """
    # Write element's tag
    fp.write_tag(elem.tag)

    # write into a buffer to avoid seeking back which can be expansive
    buffer = DicomBytesIO()
    buffer.is_little_endian = fp.is_little_endian
    buffer.is_implicit_VR = fp.is_implicit_VR

    vr: str | None = elem.VR
    if not fp.is_implicit_VR and vr and len(vr) != 2:
        msg = (
            f"Cannot write ambiguous VR of '{vr}' for data element with "
            f"tag {elem.tag!r}.\nSet the correct VR before "
            f"writing, or use an implicit VR transfer syntax"
        )
        raise ValueError(msg)

    if elem.is_raw:
        elem = cast(RawDataElement, elem)
        # raw data element values can be written as they are
        buffer.write(cast(bytes, elem.value))
        is_undefined_length = elem.length == 0xFFFFFFFF
    else:
        elem = cast(DataElement, elem)
        if vr not in writers:
            raise NotImplementedError(
                f"write_data_element: unknown Value Representation '{vr}'"
            )

        encodings = encodings or [default_encoding]
        encodings = convert_encodings(encodings)
        fn, param = writers[cast(VR, vr)]
        is_undefined_length = elem.is_undefined_length
        if not elem.is_empty:
            if vr in CUSTOMIZABLE_CHARSET_VR or vr == VR.SQ:
                fn(buffer, elem, encodings=encodings)  # type: ignore[operator]
            else:
                # Many numeric types use the same writer but with
                # numeric format parameter
                if param is not None:
                    fn(buffer, elem, param)  # type: ignore[operator]
                elif not elem.is_buffered:
                    # defer writing a buffered value until after we have written the
                    # tag and length in the file
                    fn(buffer, elem)  # type: ignore[operator]

    # valid pixel data with undefined length shall contain encapsulated
    # data, e.g. sequence items - raise ValueError otherwise (see #238)
    if is_undefined_length and elem.tag == 0x7FE00010:
        if elem.is_buffered:
            value = cast(BufferedIOBase, elem.value)
            with reset_buffer_position(value):
                pixel_data_bytes = value.read(4)
        else:
            pixel_data_bytes = cast(bytes, elem.value)[:4]

        # Big endian encapsulation is non-conformant
        tag = b"\xFE\xFF\x00\xE0" if fp.is_little_endian else b"\xFF\xFE\xE0\x00"
        if not pixel_data_bytes.startswith(tag):
            raise ValueError(
                "The (7FE0,0010) 'Pixel Data' element value hasn't been "
                "encapsulated as required for a compressed transfer syntax - "
                "see pydicom.encaps.encapsulate() for more information"
            )

    value_length = (
        buffer.tell()
        if not elem.is_buffered
        else buffer_remaining(cast(BufferedIOBase, elem.value))
    )

    if (
        not fp.is_implicit_VR
        and vr not in EXPLICIT_VR_LENGTH_32
        and not is_undefined_length
        and value_length > 0xFFFF
    ):
        # see PS 3.5, section 6.2.2 for handling of this case
        warn_and_log(
            f"The value for the data element {elem.tag} exceeds the "
            f"size of 64 kByte and cannot be written in an explicit transfer "
            f"syntax. The data element VR is changed from '{vr}' to 'UN' "
            f"to allow saving the data."
        )
        vr = VR.UN

    # write the VR for explicit transfer syntax
    if not fp.is_implicit_VR:
        vr = cast(str, vr)
        fp.write(bytes(vr, default_encoding))

        if vr in EXPLICIT_VR_LENGTH_32:
            fp.write_US(0)  # reserved 2 bytes

    if (
        not fp.is_implicit_VR
        and vr not in EXPLICIT_VR_LENGTH_32
        and not is_undefined_length
    ):
        fp.write_US(value_length)  # Explicit VR length field is 2 bytes
    else:
        # write the proper length of the data_element in the length slot,
        # unless is SQ with undefined length.
        fp.write_UL(0xFFFFFFFF if is_undefined_length else value_length)

    # if the value is buffered, now we want to write the value directly to the fp
    if elem.is_buffered:
        fn(fp, elem)  # type: ignore[operator]
    else:
        fp.write(buffer.getvalue())

    if is_undefined_length:
        fp.write_tag(SequenceDelimiterTag)
        fp.write_UL(0)  # 4-byte 'length' of delimiter data item


EncodingType = tuple[bool | None, bool | None]


def write_dataset(
    fp: DicomIO, dataset: Dataset, parent_encoding: str | list[str] = default_encoding
) -> int:
    """Encode `dataset` and write the encoded data to `fp`.

    *Encoding*

    The `dataset` is encoded as specified by (in order of priority):

    * ``fp.is_implicit_VR`` and ``fp.is_little_endian``.
    * ``dataset.is_implicit_VR`` and ``dataset.is_little_endian``
    * If `dataset` has been decoded from a file or buffer then
      :attr:`~pydicom.dataset.Dataset.original_encoding`.

    Parameters
    ----------
    fp : pydicom.filebase.DicomIO
        The file-like to write the encoded dataset to.
    dataset : pydicom.dataset.Dataset
        The dataset to be encoded.
    parent_encoding : str | List[str], optional
        The character set to use for encoding strings, defaults to ``"iso8859"``.

    Returns
    -------
    int
        The number of bytes written to `fp`.
    """
    # TODO: Update in v4.0
    # In order of encoding priority

    fp_encoding: EncodingType = (
        getattr(fp, "_implicit_vr", None),
        getattr(fp, "_little_endian", None),
    )
    ds_encoding: EncodingType = (None, None)
    if not config._use_future:
        ds_encoding = (dataset.is_implicit_VR, dataset.is_little_endian)
    or_encoding = dataset.original_encoding

    if None in fp_encoding and None in ds_encoding and None in or_encoding:
        raise AttributeError(
            "'fp.is_implicit_VR' and 'fp.is_little_endian' attributes are required "
        )

    if None in fp_encoding:
        if None not in ds_encoding:
            fp_encoding = ds_encoding
        elif None not in or_encoding:
            fp_encoding = or_encoding

    fp.is_implicit_VR, fp.is_little_endian = cast(tuple[bool, bool], fp_encoding)
    get_item: Callable[[BaseTag], DataElement | RawDataElement] = dataset.get_item

    # This function is doing some heavy lifting:
    #   If implicit -> explicit, runs ambiguous VR correction
    #   If implicit -> explicit, RawDataElements -> DataElement (VR lookup)
    #   If charset changed, RawDataElements -> DataElement
    if (
        fp_encoding != or_encoding
        or dataset.original_character_set != dataset._character_set
    ):
        dataset = correct_ambiguous_vr(dataset, fp.is_little_endian)
        # Use __getitem__ instead or get_item to force parsing of RawDataElements into DataElements,
        # so we can re-encode them with the correct charset and encoding
        get_item = dataset.__getitem__

    dataset_encoding = cast(
        None | str | list[str], dataset.get("SpecificCharacterSet", parent_encoding)
    )

    fpStart = fp.tell()

    # data_elements must be written in tag order
    for tag in sorted(dataset.keys()):
        # do not write retired Group Length (see PS3.5, 7.2)
        if tag.element == 0 and tag.group > 6:
            continue

        with tag_in_exception(tag):
            write_data_element(fp, get_item(tag), dataset_encoding)

    return fp.tell() - fpStart


def write_sequence(fp: DicomIO, elem: DataElement, encodings: list[str]) -> None:
    """Write a sequence contained in `data_element` to the file-like `fp`.

    Parameters
    ----------
    fp : file-like
        The file-like to write the encoded data to.
    data_element : dataelem.DataElement
        The sequence element to write to `fp`.
    encodings : list of str
        The character encodings to use on text values.
    """
    # write_data_element has already written the VR='SQ' (if needed) and
    #    a placeholder for length"""
    for ds in cast(Iterable[Dataset], elem.value):
        write_sequence_item(fp, ds, encodings)


def write_sequence_item(fp: DicomIO, dataset: Dataset, encodings: list[str]) -> None:
    """Write a `dataset` in a sequence to the file-like `fp`.

    This is similar to writing a data_element, but with a specific tag for
    Sequence Item.

    See DICOM Standard, Part 5, :dcm:`Section 7.5<sect_7.5.html>`.

    Parameters
    ----------
    fp : file-like
        The file-like to write the encoded data to.
    dataset : Dataset
        The :class:`Dataset<pydicom.dataset.Dataset>` to write to `fp`.
    encodings : list of str
        The character encodings to use on text values.
    """
    fp.write_tag(ItemTag)  # marker for start of Sequence Item
    length_location = fp.tell()  # save location for later.
    # will fill in real value later if not undefined length
    fp.write_UL(0xFFFFFFFF)
    write_dataset(fp, dataset, parent_encoding=encodings)
    if getattr(dataset, "is_undefined_length_sequence_item", False):
        fp.write_tag(ItemDelimiterTag)
        fp.write_UL(0)  # 4-bytes 'length' field for delimiter item
    else:  # we will be nice and set the lengths for the reader of this file
        location = fp.tell()
        fp.seek(length_location)
        fp.write_UL(location - length_location - 4)  # 4 is length of UL
        fp.seek(location)  # ready for next data_element


def write_UN(fp: DicomIO, elem: DataElement) -> None:
    """Write a byte string for an DataElement of value 'UN' (unknown)."""
    fp.write(cast(bytes, elem.value))


def write_ATvalue(fp: DicomIO, elem: DataElement) -> None:
    """Write a data_element tag to a file."""
    try:
        iter(cast(Sequence[Any], elem.value))  # see if is multi-valued AT;
        # Note will fail if Tag ever derived from true tuple rather than being
        # a long
    except TypeError:
        # make sure is expressed as a Tag instance
        tag = Tag(cast(int, elem.value))
        fp.write_tag(tag)
    else:
        tags = [Tag(tag) for tag in cast(Sequence[int], elem.value)]
        for tag in tags:
            fp.write_tag(tag)


def write_file_meta_info(
    fp: DicomIO, file_meta: FileMetaDataset, enforce_standard: bool = True
) -> None:
    """Write the File Meta Information elements in `file_meta` to `fp`.

    If `enforce_standard` is ``True`` then the file-like `fp` should be
    positioned past the 128 byte preamble + 4 byte prefix (which should
    already have been written).

    **DICOM File Meta Information Group Elements**

    From the DICOM standard, Part 10,
    :dcm:`Section 7.1<part10/chapter_7.html#sect_7.1>`,  any DICOM file shall
    contain a 128-byte preamble, a 4-byte DICOM prefix 'DICM' and (at a
    minimum) the following Type 1 DICOM Elements (from
    :dcm:`Table 7.1-1<part10/chapter_7.html#table_7.1-1>`):

    * (0002,0000) *File Meta Information Group Length*, UL, 4
    * (0002,0001) *File Meta Information Version*, OB, 2
    * (0002,0002) *Media Storage SOP Class UID*, UI, N
    * (0002,0003) *Media Storage SOP Instance UID*, UI, N
    * (0002,0010) *Transfer Syntax UID*, UI, N
    * (0002,0012) *Implementation Class UID*, UI, N

    If `enforce_standard` is ``True`` then (0002,0000) will be added/updated,
    (0002,0001) and (0002,0012) will be added if not already present and the
    other required elements will be checked to see if they exist. If
    `enforce_standard` is ``False`` then `file_meta` will be written as is
    after minimal validation checking.

    The following Type 3/1C Elements may also be present:

    * (0002,0013) *Implementation Version Name*, SH, N
    * (0002,0016) *Source Application Entity Title*, AE, N
    * (0002,0017) *Sending Application Entity Title*, AE, N
    * (0002,0018) *Receiving Application Entity Title*, AE, N
    * (0002,0102) *Private Information*, OB, N
    * (0002,0100) *Private Information Creator UID*, UI, N

    If `enforce_standard` is ``True`` then (0002,0013) will be added/updated.

    *Encoding*

    The encoding of the *File Meta Information* shall be *Explicit VR Little
    Endian*.

    Parameters
    ----------
    fp : file-like
        The file-like to write the File Meta Information to.
    file_meta : pydicom.dataset.Dataset
        The File Meta Information elements.
    enforce_standard : bool
        If ``False``, then only the *File Meta Information* elements already in
        `file_meta` will be written to `fp`. If ``True`` (default) then a DICOM
        Standards conformant File Meta will be written to `fp`.

    Raises
    ------
    ValueError
        If `enforce_standard` is ``True`` and any of the required *File Meta
        Information* elements are missing from `file_meta`, with the
        exception of (0002,0000), (0002,0001) and (0002,0012).
    ValueError
        If any non-Group 2 Elements are present in `file_meta`.
    """
    validate_file_meta(file_meta, enforce_standard)

    if enforce_standard and "FileMetaInformationGroupLength" not in file_meta:
        # Will be updated with the actual length later
        file_meta.FileMetaInformationGroupLength = 0

    # Write the File Meta Information Group elements
    # first write into a buffer to avoid seeking back, that can be
    # expansive and is not allowed if writing into a zip file
    buffer = DicomBytesIO()
    buffer.is_little_endian = True
    buffer.is_implicit_VR = False
    write_dataset(buffer, file_meta)

    # If FileMetaInformationGroupLength is present it will be the first written
    #   element and we must update its value to the correct length.
    if "FileMetaInformationGroupLength" in file_meta:
        # Update the FileMetaInformationGroupLength value, which is the number
        #   of bytes from the end of the FileMetaInformationGroupLength element
        #   to the end of all the File Meta Information elements.
        # FileMetaInformationGroupLength has a VR of 'UL' and so has a value
        #   that is 4 bytes fixed. The total length of when encoded as
        #   Explicit VR must therefore be 12 bytes.
        file_meta.FileMetaInformationGroupLength = buffer.tell() - 12
        buffer.seek(0)
        write_data_element(buffer, file_meta[0x00020000])

    fp.write(buffer.getvalue())


def _determine_encoding(
    ds: Dataset,
    tsyntax: UID | None,
    implicit_vr: bool | None,
    little_endian: bool | None,
    force_encoding: bool,
) -> tuple[bool, bool]:
    """Return the encoding to use for `ds`.

    If `force_encoding` isn't ``True`` the priority is:

    1. The encoding corresponding to `tsyntax`
    2. The encoding set by `implicit_vr` and `little_endian`
    3. `Dataset.is_implicit_VR` and `Dataset.is_little_endian`
    4. `Dataset.original_encoding`

    If none of those are valid, raise an exception.

    If `force_encoding` is ``True`` then `implicit_vr` and `little_endian` are
    required.

    Parameters
    ----------
    ds : pydicom.dataset.Dataset
        The dataset that is to be encoded.
    tsyntax : pydicom.uid.UID | None
        The dataset's public or private transfer syntax (if any). Private
        transfer syntaxes require `implicit_vr` and `little_endian` be used.
    implicit_vr : bool | None
        The VR encoding method (if supplied)
    little_endian : bool | None
        The encoding endianness (if supplied).
    force_encoding : bool
        If ``True`` then force the encoding to use `implicit_vr` and
        `little_endian`. Default ``False``.

    Returns
    -------
    tuple[bool, bool]
        The encoding to use as ``[use implicit VR, use little endian]``.

    Raises
    ------
    ValueError
        If unable to determine the encoding to use, if `transfer_syntax` is
        not a transfer syntax, or if there's an inconsistency between
        `transfer_syntax` and `implicit_vr` or `little_endian`.
    """
    arg_encoding = (implicit_vr, little_endian)
    if force_encoding:
        if None in arg_encoding:
            raise ValueError(
                "'implicit_vr' and 'little_endian' are required if "
                "'force_encoding' is used"
            )

        return cast(tuple[bool, bool], arg_encoding)

    # The default for little_endian is `None` so we can require the use of
    #   args with `force_encoding`, but we actually default it to `True`
    #   when `implicit_vr` is used as a fallback
    if implicit_vr is not None and little_endian is None:
        arg_encoding = (implicit_vr, True)

    ds_encoding: EncodingType = (None, None)
    if not config._use_future:
        ds_encoding = (ds.is_implicit_VR, ds.is_little_endian)

    fallback_encoding: EncodingType = (None, None)
    if None not in arg_encoding:
        fallback_encoding = arg_encoding
    elif None not in ds_encoding:
        fallback_encoding = ds_encoding
    elif None not in ds.original_encoding:
        fallback_encoding = ds.original_encoding

    if tsyntax is None:
        if None not in fallback_encoding:
            return cast(tuple[bool, bool], fallback_encoding)

        raise ValueError(
            "Unable to determine the encoding to use for writing the dataset, "
            "please set the file meta's Transfer Syntax UID or use the "
            "'implicit_vr' and 'little_endian' arguments"
        )

    if tsyntax.is_private and not tsyntax.is_transfer_syntax:
        if None in fallback_encoding:
            raise ValueError(
                "The 'implicit_vr' and 'little_endian' arguments are required "
                "when using a private transfer syntax"
            )

        return cast(tuple[bool, bool], fallback_encoding)

    if not tsyntax.is_transfer_syntax:
        raise ValueError(
            f"The Transfer Syntax UID '{tsyntax.name}' is not a valid "
            "transfer syntax"
        )

    # Check that supplied args match transfer syntax
    if implicit_vr is not None and implicit_vr != tsyntax.is_implicit_VR:
        raise ValueError(
            f"The 'implicit_vr' value is not consistent with the required "
            f"VR encoding for the '{tsyntax.name}' transfer syntax"
        )

    if little_endian is not None and little_endian != tsyntax.is_little_endian:
        raise ValueError(
            f"The 'little_endian' value is not consistent with the required "
            f"endianness for the '{tsyntax.name}' transfer syntax"
        )

    return (tsyntax.is_implicit_VR, tsyntax.is_little_endian)


def dcmwrite(
    filename: PathType | BinaryIO | WriteableBuffer,
    dataset: Dataset,
    /,
    __write_like_original: bool | None = None,
    *,
    implicit_vr: bool | None = None,
    little_endian: bool | None = None,
    enforce_file_format: bool = False,
    force_encoding: bool = False,
    overwrite: bool = True,
    **kwargs: Any,
) -> None:
    """Write `dataset` to `filename`, which can be a path, a file-like or a
    writeable buffer.

    .. versionchanged:: 3.0

        Added the `enforce_file_format` and `overwrite` keyword arguments.

    .. deprecated:: 3.0

        `write_like_original` is deprecated and will be removed in v4.0, use
        `enforce_file_format` instead.

    If `enforce_file_format` is ``True`` then an attempt will be made to write
    `dataset` using the :dcm:`DICOM File Format <part10/chapter_7.html>`, or
    raise an exception if unable to do so.

    If `enforce_file_format` is ``False`` (default) then `dataset` will be
    written as-is (after minimal validation checking) and may or may not
    contain all or parts of the *File Meta Information* and hence may or may
    not be conformant with the DICOM File Format.

    **DICOM File Format**

    The *DICOM File Format* consists of a 128-byte preamble, a 4 byte
    ``b'DICM'`` prefix, the *File Meta Information Group* elements and finally
    the encoded `dataset`.

    **Preamble and Prefix**

    The ``dataset.preamble`` attribute shall be 128-bytes long or ``None``. The
    actual preamble written depends on `enforce_file_format` and
    ``dataset.preamble`` (see the table below).

    +------------------+------------------------------+
    |                  | enforce_file_format          |
    +------------------+-------------+----------------+
    | dataset.preamble | False       | True           |
    +==================+=============+================+
    | None             | no preamble | 128 0x00 bytes |
    +------------------+-------------+----------------+
    | 128 bytes        | dataset.preamble             |
    +------------------+------------------------------+

    The prefix shall be the bytestring ``b'DICM'`` and will be written if and
    only if the preamble is present.

    **File Meta Information Group Elements**

    The preamble and prefix are followed by a set of DICOM elements from the
    (0002,eeee) group. Some of these elements are required (Type 1) while
    others are optional (Type 3/1C). If `enforce_file_format` is ``False``
    then the *File Meta Information Group* elements are all optional, otherwise
    an attempt will be made to add the required elements using `dataset`. See
    :func:`~pydicom.filewriter.write_file_meta_info` for more information on
    which elements are required.

    The *File Meta Information Group* elements must be included within their
    own :class:`~pydicom.dataset.FileMetaDataset` in the ``dataset.file_meta``
    attribute.

    *Encoding*

    The preamble and prefix are encoding independent. The *File Meta
    Information Group* elements are encoded as *Explicit VR Little Endian* as
    required by the DICOM Standard.

    **Dataset**

    A DICOM Dataset representing a SOP Instance related to a DICOM Information
    Object Definition (IOD). It's up to the user to ensure `dataset` conforms
    to the requirements of the IOD.

    *Encoding*

    .. versionchanged:: 3.0

        Added the `implicit_vr` and `little_endian` arguments.

    The `dataset` is encoded as specified by (in order of priority):

    * The encoding corresponding to the set *Transfer Syntax UID* in
      :attr:`~pydicom.dataset.FileDataset.file_meta`.
    * The `implicit_vr` and `little_endian` arguments
    * :attr:`~pydicom.dataset.Dataset.is_implicit_VR` and
      :attr:`~pydicom.dataset.Dataset.is_little_endian`
    * :attr:`~pydicom.dataset.Dataset.original_encoding`

    .. warning::

        This function does not automatically convert `dataset` from little
        to big endian encoding (or vice versa). The endianness of values for
        elements with a VR of **OD**, **OF**, **OL**, **OW**, **OV** and
        **UN** must be converted manually prior to calling
        :func:`~pydicom.filewriter.dcmwrite`.

    Parameters
    ----------
    filename : str, PathLike, file-like or writeable buffer
        File path, file-like or writeable buffer to write the encoded `dataset`
        to. If using a writeable buffer it must have ``write()``, ``seek()``
        and ``tell()`` methods.
    dataset : pydicom.dataset.FileDataset
        The dataset to be encoded.
    write_like_original : bool, optional
        If ``True`` (default) then write `dataset` as-is, otherwise
        ensure that `dataset` is written in the DICOM File Format or
        raise an exception if that isn't possible. This parameter is
        deprecated, please use `enforce_file_format` instead.
    implicit_vr : bool, optional
        Required if `dataset` has no valid public *Transfer Syntax UID*
        set in the file meta and `dataset` has been created from scratch. If
        ``True`` then encode `dataset` using implicit VR, otherwise use
        explicit VR.
    little_endian : bool, optional
        Required if `dataset` has no valid public *Transfer Syntax UID*
        set in the file meta and `dataset` has been created from scratch. If
        ``True`` (default) then use little endian byte order when encoding
        `dataset`, otherwise use big endian.
    enforce_file_format : bool, optional
        If ``True`` then ensure `dataset` is written in the DICOM File
        Format or raise an exception if that isn't possible.

        If ``False`` (default) then write `dataset` as-is, preserving the
        following - which may result in a non-conformant file:

        - ``dataset.preamble``: if `dataset` has no preamble then none will
          be written
        - ``dataset.file_meta``: if `dataset` is missing any required *File
          Meta Information Group* elements then they will not be written.
    force_encoding : bool, optional
        If ``True`` then force the encoding to follow `implicit_vr` and
        `little_endian`. Cannot be used with `enforce_file_format`. Default
        ``False``.
    overwrite : bool, optional
        If ``False`` and `filename` is a :class:`str` or PathLike, then raise a
        :class:`FileExistsError` if a file already exists with the given filename
        (default ``True``).

    Raises
    ------
    ValueError

      * If group ``0x0000`` *Command Set* elements are present in `dataset`.
      * If group ``0x0002`` *File Meta Information Group* elements are present
        in `dataset`.
      * If ``dataset.preamble`` exists but is not 128 bytes long.

    See Also
    --------
    pydicom.dataset.Dataset
        Dataset class with relevant attributes and information.
    pydicom.dataset.Dataset.save_as
        Encode a dataset and write it to file, wraps ``dcmwrite()``.
    """
    # TODO: Remove in v4.0
    # Cover use of `write_like_original` as:
    #   optional arg - dcmwrite(fp, ds, write_like_original=bool)
    #   positional arg - dcmwrite(fp, ds, False)
    write_like_original: bool | None = kwargs.get("write_like_original", None)
    if None not in (__write_like_original, write_like_original):
        if config._use_future:
            raise TypeError(
                "'write_like_original' is no longer accepted as a positional "
                "or keyword argument, use 'enforce_file_format' instead"
            )

        raise TypeError(
            "'write_like_original' cannot be used as both a positional "
            "and keyword argument"
        )

    if write_like_original is None:
        write_like_original = __write_like_original

    if write_like_original is not None:
        if config._use_future:
            raise TypeError(
                "'write_like_original' is no longer accepted as a positional "
                "or keyword argument, use 'enforce_file_format' instead"
            )

        warn_and_log(
            (
                "'write_like_original' is deprecated and will be removed in "
                "v4.0, please use 'enforce_file_format' instead"
            ),
            DeprecationWarning,
        )
        enforce_file_format = not write_like_original

    # Ensure kwargs only contains `write_like_original`
    keys = [x for x in kwargs.keys() if x != "write_like_original"]
    if keys:
        raise TypeError(
            f"Invalid keyword argument(s) for dcmwrite(): {', '.join(keys)}"
        )

    cls_name = dataset.__class__.__name__

    # Check for disallowed tags
    bad_tags = [x >> 16 for x in dataset._dict if x >> 16 in (0, 2)]
    if bad_tags:
        if 0 in bad_tags:
            raise ValueError(
                "Command Set elements (0000,eeee) are not allowed when using "
                "dcmwrite(), use write_dataset() instead"
            )
        else:
            raise ValueError(
                "File Meta Information Group elements (0002,eeee) must be in a "
                f"FileMetaDataset instance in the '{cls_name}.file_meta' attribute"
            )

    if force_encoding and enforce_file_format:
        raise ValueError("'force_encoding' cannot be used with 'enforce_file_format'")

    # Avoid making changes to the original File Meta Information
    file_meta = FileMetaDataset()
    if hasattr(dataset, "file_meta"):
        file_meta = deepcopy(dataset.file_meta)

    tsyntax: UID | None = file_meta.get("TransferSyntaxUID", None)

    # The dataset encoding method
    encoding = _determine_encoding(
        dataset,
        tsyntax,
        implicit_vr,
        little_endian,
        force_encoding,
    )

    if not force_encoding and encoding == (True, False):
        raise ValueError(
            "Implicit VR and big endian is not a valid encoding combination"
        )

    preamble = getattr(dataset, "preamble", None)
    if preamble and len(preamble) != 128:
        raise ValueError(f"'{cls_name}.preamble' must be 128-bytes long")

    if enforce_file_format:
        # A valid File Meta Information is required
        if tsyntax is None:
            if encoding == (True, True):
                file_meta.TransferSyntaxUID = ImplicitVRLittleEndian
            elif encoding == (False, False):
                file_meta.TransferSyntaxUID = ExplicitVRBigEndian

            tsyntax = file_meta.get("TransferSyntaxUID", None)

        # Ensure the file_meta Class and Instance UIDs are up to date
        #   but don't overwrite if nothing is set in the dataset
        meta_class = file_meta.get("MediaStorageSOPClassUID", None)
        ds_class = dataset.get("SOPClassUID", None)
        if meta_class is None or (ds_class and ds_class != meta_class):
            file_meta.MediaStorageSOPClassUID = ds_class

        meta_instance = file_meta.get("MediaStorageSOPInstanceUID", None)
        ds_instance = dataset.get("SOPInstanceUID", None)
        if meta_instance is None or (ds_instance and ds_instance != meta_instance):
            file_meta.MediaStorageSOPInstanceUID = ds_instance

        # Will raise if the file meta isn't valid
        validate_file_meta(file_meta, enforce_standard=True)

        # A preamble is required
        if not preamble:
            preamble = b"\x00" * 128

    if tsyntax and not tsyntax.is_private and tsyntax.is_transfer_syntax:
        # PS3.5 Annex A.4 - the length of encapsulated pixel data is undefined
        #   and native pixel data uses actual length
        if "PixelData" in dataset:
            dataset["PixelData"].is_undefined_length = tsyntax.is_compressed

    caller_owns_file = True
    # Open file if not already a file object
    filename = path_from_pathlike(filename)
    if isinstance(filename, str):
        # A path-like to be written to
        file_mode = "xb" if not overwrite else "wb"
        fp: DicomIO = DicomFile(filename, file_mode)
        # caller provided a file name; we own the file handle
        caller_owns_file = False
    elif isinstance(filename, DicomIO):
        # A wrapped writeable buffer, don't wrap it again
        fp = filename
    else:
        # Anything else
        try:
            fp = DicomIO(filename)
        except AttributeError as exc:
            raise TypeError(
                "dcmwrite: Expected a file path, file-like or writeable buffer, "
                f"but got {type(filename).__name__}"
            ) from exc

    # Set the encoding of the buffer/file-like
    fp.is_implicit_VR, fp.is_little_endian = encoding

    try:
        if preamble:
            # Write the 'DICM' prefix if and only if we write the preamble
            fp.write(preamble)
            fp.write(b"DICM")

        if file_meta:  # May be empty
            write_file_meta_info(fp, file_meta, enforce_standard=enforce_file_format)

        if tsyntax == DeflatedExplicitVRLittleEndian:
            # See PS3.5 section A.5
            # When writing, the entire dataset following the file meta data
            #   is encoded normally, then "deflate" compression applied
            buffer = DicomBytesIO()
            buffer.is_implicit_VR, buffer.is_little_endian = encoding
            write_dataset(buffer, dataset)

            # Compress the encoded data and write to file
            compressor = zlib.compressobj(wbits=-zlib.MAX_WBITS)
            deflated = bytearray(compressor.compress(buffer.getvalue()))
            deflated += compressor.flush()
            fp.write(deflated)
            if len(deflated) % 2:
                fp.write(b"\x00")

        else:
            write_dataset(fp, dataset)

    finally:
        if not caller_owns_file:
            fp.close()


# Map each VR to a function which can write it
# for write_numbers, the Writer maps to a tuple (function, struct_format)
#   (struct_format is python's struct module format)
writers = {
    VR.AE: (write_string, None),
    VR.AS: (write_string, None),
    VR.AT: (write_ATvalue, None),
    VR.CS: (write_string, None),
    VR.DA: (write_DA, None),
    VR.DS: (write_number_string, None),
    VR.DT: (write_DT, None),
    VR.FD: (write_numbers, "d"),
    VR.FL: (write_numbers, "f"),
    VR.IS: (write_number_string, None),
    VR.LO: (write_text, None),
    VR.LT: (write_text, None),
    VR.OB: (write_OBvalue, None),
    VR.OD: (write_OWvalue, None),
    VR.OF: (write_OWvalue, None),
    VR.OL: (write_OWvalue, None),
    VR.OW: (write_OWvalue, None),
    VR.OV: (write_OWvalue, None),
    VR.PN: (write_PN, None),
    VR.SH: (write_text, None),
    VR.SL: (write_numbers, "l"),
    VR.SQ: (write_sequence, None),
    VR.SS: (write_numbers, "h"),
    VR.ST: (write_text, None),
    VR.SV: (write_numbers, "q"),
    VR.TM: (write_TM, None),
    VR.UC: (write_text, None),
    VR.UI: (write_UI, None),
    VR.UL: (write_numbers, "L"),
    VR.UN: (write_UN, None),
    VR.UR: (write_string, None),
    VR.US: (write_numbers, "H"),
    VR.UT: (write_text, None),
    VR.UV: (write_numbers, "Q"),
    VR.US_SS: (write_OWvalue, None),
    VR.US_OW: (write_OWvalue, None),
    VR.US_SS_OW: (write_OWvalue, None),
    VR.OB_OW: (write_OBvalue, None),
}



# Copyright 2008-2024 pydicom authors. See LICENSE file for details.
"""Pixel data encoding."""

from collections.abc import Callable, Iterator, Iterable
import logging
import math
import sys
from typing import Any, cast, TYPE_CHECKING

try:
    import numpy as np
except ImportError:
    pass

from pydicom import config
from pydicom.pixels.common import Buffer, RunnerBase, CoderBase, RunnerOptions
from pydicom.uid import (
    UID,
    JPEGBaseline8Bit,
    JPEGExtended12Bit,
    JPEGLossless,
    JPEGLosslessSV1,
    JPEGLSLossless,
    JPEGLSNearLossless,
    JPEG2000Lossless,
    JPEG2000,
    RLELossless,
    JPEGLSTransferSyntaxes,
)

if TYPE_CHECKING:  # pragma: no cover
    from pydicom.dataset import Dataset


LOGGER = logging.getLogger(__name__)


EncodeFunction = Callable[[bytes, "EncodeRunner"], bytes | bytearray]


class EncodeOptions(RunnerOptions, total=False):
    """Options accepted by EncodeRunner and encoding plugins"""

    # The byte order used by the raw pixel data sent to the encoder
    #   ">" for big endian, "<" for little endian (default)
    byteorder: str

    ## Transfer Syntax specific options
    # JPEG-LS Near Lossless
    # The maximum allowable error in *unsigned* pixel intensity
    jls_error: int

    # JPEG 2000
    # Either j2k_cr or j2k_psnr is required
    # The compression ratio for each quality layer, should be in decreasing order
    j2k_cr: list[float]
    # The peak signal-to-noise ratio for each layer, should be in increasing order
    j2k_psnr: list[float]


class EncodeRunner(RunnerBase):
    """Class for managing the pixel data encoding process.

    .. versionadded:: 3.0

    This class is not intended to be used directly. For encoding pixel data
    use the :class:`~pydicom.pixels.encoders.base.Encoder` instance
    corresponding to the transfer syntax of the pixel data.
    """

    def __init__(self, tsyntax: UID) -> None:
        """Create a new runner for encoding data as `tsyntax`.

        Parameters
        ----------
        tsyntax : pydicom.uid.UID
            The transfer syntax UID corresponding to the pixel data to be
            encoded.
        """
        self._src: Buffer | np.ndarray
        self._src_type: str
        self._opts: EncodeOptions = {
            "transfer_syntax_uid": tsyntax,
            "byteorder": "<",
            "pixel_keyword": "PixelData",
        }
        self._undeletable = ("transfer_syntax_uid", "pixel_keyword", "byteorder")
        self._encoders: dict[str, EncodeFunction] = {}

    def encode(self, index: int | None) -> bytes:
        """Return an encoded frame of pixel data as :class:`bytes`.

        Parameters
        ----------
        index : int | None
            If `index` is ``None`` then the pixel data only contains one frame,
            otherwise `index` is the frame number to be encoded.

        Returns
        ------
        bytes
            The encoded pixel data frame.
        """
        failure_messages = []
        for name, func in self._encoders.items():
            try:
                return func(self.get_frame(index), self)
            except Exception as exc:
                LOGGER.exception(exc)
                failure_messages.append(f"{name}: {exc}")

        messages = "\n  ".join(failure_messages)
        raise RuntimeError(
            "Unable to encode as exceptions were raised by all available "
            f"plugins:\n  {messages}"
        )

    def get_frame(self, index: int | None) -> bytes:
        """Return a frame's worth of uncompressed pixel data as :class:`bytes`.

        Parameters
        ----------
        index : int | None
            If the pixel data only has one from then use ``None``, otherwise
            `index` is the index of the frame to be returned.
        """
        if self.is_array:
            return self._get_frame_array(index)

        frame = self._get_frame_buffer(index)
        return bytes(frame) if not isinstance(frame, bytes) else frame

    def _get_frame_array(self, index: int | None) -> bytes:
        """Return a frame's worth of uncompressed pixel data from an ndarray."""
        # Grab the frame so that subsequent array manipulations minimize
        #   the memory usage
        arr = cast(np.ndarray, self.src[index] if index is not None else self.src)

        # The ndarray containing the pixel data may use a larger container
        #   than is strictly needed: e.g. 32 bits allocated with 7 bits stored.
        #   However the encoders expect data to be sized appropriately
        #   for the sample precision, so we may need to downscale
        if self.bits_stored <= 8:
            itemsize = 1
        elif 8 < self.bits_stored <= 16:
            itemsize = 2
        elif 16 < self.bits_stored <= 32:
            itemsize = 4
        elif 32 < self.bits_stored <= 64:
            itemsize = 8

        if arr.dtype.itemsize != itemsize:
            arr = arr.astype(f"{arr.dtype.kind}{itemsize}")

        # JPEG-LS allows different ordering of the input image data via the
        #   interleave mode (ILV) parameter. ILV 0 matches a planar configuration
        #   of 1 and requires shape (samples, rows, columns). ILV 1 and 2 match
        #   planar configuration 0 so no further action is needed
        if (
            self.transfer_syntax in JPEGLSTransferSyntaxes
            and self.samples_per_pixel == 3
            and self.planar_configuration == 1
        ):
            arr = arr.transpose(2, 0, 1)

        return cast(bytes, arr.tobytes())

    def _get_frame_buffer(self, index: int | None) -> bytes | bytearray:
        """Return a frame's worth of uncompressed pixel data from buffer-like."""
        # The encoded pixel data may use a larger container than is strictly
        #   needed: e.g. 32 bits allocated with 7 bits stored
        # However the encoders typically expect data to be sized appropriately
        # for the sample precision, so we need to downscale to:
        #    0 < precision <=  8: an 8-bit container (char)
        #    8 < precision <= 16: a 16-bit container (short)
        #   16 < precision <= 32: a 32-bit container (int/long)
        #   32 < precision <= 64: a 64-bit container (long long)
        bytes_per_frame = cast(int, self.frame_length(unit="bytes"))
        start = 0 if index is None else index * bytes_per_frame
        src = cast(bytes, self.src[start : start + bytes_per_frame])

        # Resize the data to fit the appropriate container
        expected_length = cast(int, self.frame_length(unit="pixels"))
        bytes_per_pixel = len(src) // expected_length

        # 1 byte/px actual
        if self.bits_stored <= 8:
            # If not 1 byte/px then must be 2, 3, 4, 5, 6, 7 or 8
            #   but only the first byte is relevant
            return src if bytes_per_pixel == 1 else src[::bytes_per_pixel]

        # 2 bytes/px actual
        if 8 < self.bits_stored <= 16:
            if bytes_per_pixel == 2:
                return src

            # If not 2 bytes/px then must be 3, 4, 5, 6, 7 or 8
            #   but only the first 2 bytes are relevant
            out = bytearray(expected_length * 2)
            out[::2] = src[::bytes_per_pixel]
            out[1::2] = src[1::bytes_per_pixel]
            return out

        # 3 or 4 bytes/px actual
        if 16 < self.bits_stored <= 32:
            if bytes_per_pixel == 4:
                return src

            # If not 4 bytes/px then must be 3, 5, 6, 7 or 8
            #   but only the first 3 or 4 bytes are relevant
            out = bytearray(expected_length * 4)
            out[::4] = src[::bytes_per_pixel]
            out[1::4] = src[1::bytes_per_pixel]
            out[2::4] = src[2::bytes_per_pixel]
            if bytes_per_pixel > 3:
                out[3::4] = src[3::bytes_per_pixel]

            return out

        # 32 < bits_stored <= 64 (maximum allowed)
        # 5, 6, 7 or 8 bytes/px actual
        if bytes_per_pixel == 8:
            return src

        # If not 8 bytes/px then must be 5, 6 or 7
        out = bytearray(expected_length * 8)
        out[::8] = src[::bytes_per_pixel]
        out[1::8] = src[1::bytes_per_pixel]
        out[2::8] = src[2::bytes_per_pixel]
        out[3::8] = src[3::bytes_per_pixel]
        out[4::8] = src[4::bytes_per_pixel]
        if bytes_per_pixel == 5:
            return out

        out[5::8] = src[5::bytes_per_pixel]
        if bytes_per_pixel == 6:
            return out

        # 7 bytes/px
        out[6::8] = src[6::bytes_per_pixel]
        return out

    def set_encoders(self, encoders: dict[str, EncodeFunction]) -> None:
        """Set the encoders use for encoding compressed pixel data.

        Parameters
        ----------
        encoders : dict[str, EncodeFunction]
            A dict of {name: encoder function}.
        """
        self._encoders = encoders

    def set_source(self, src: "np.ndarray | Dataset | Buffer") -> None:
        """Set the pixel data to be encoded.

        Parameters
        ----------
        src : bytes | bytearray | memoryview | pydicom.dataset.Dataset | numpy.ndarray

            * If a buffer-like then the encoded pixel data
            * If a :class:`~pydicom.dataset.Dataset` then a dataset containing
              the pixel data and associated group ``0x0028`` elements.
            * If a :class:`numpy.ndarray` then an array containing the image data.
        """
        from pydicom.dataset import Dataset

        if isinstance(src, Dataset):
            self._set_options_ds(src)
            self._src = src.PixelData
            self._src_type = "Dataset"
        elif isinstance(src, (bytes | bytearray | memoryview)):
            self._src = src
            self._src_type = "Buffer"
        elif isinstance(src, np.ndarray):
            # Ensure the array is in the required byte order (little-endian)
            sys_endianness = "<" if sys.byteorder == "little" else ">"
            # `byteorder` may be
            #   '|': none available, such as for 8 bit -> ignore
            #   '=': native system endianness -> change to '<' or '>'
            #   '<' or '>': little or big
            byteorder = src.dtype.byteorder
            byteorder = sys_endianness if byteorder == "=" else byteorder
            if byteorder == ">":
                src = src.astype(src.dtype.newbyteorder("<"))

            self._src = src
            self._src_type = "Array"
        else:
            raise TypeError(
                "'src' must be bytes, numpy.ndarray or pydicom.dataset.Dataset, "
                f"not '{src.__class__.__name__}'"
            )

    @property
    def src(self) -> "Buffer | np.ndarray":
        """Return the buffer-like or :class:`numpy.ndarray` containing the pixel data."""
        return self._src

    def __str__(self) -> str:
        """Return nice string output for the runner."""
        s = [f"EncodeRunner for '{self.transfer_syntax.name}'"]
        s.append("Options")
        s.extend([f"  {name}: {value}" for name, value in self.options.items()])
        if self._encoders:
            s.append("Encoders")
            s.extend([f"  {name}" for name in self._encoders])

        return "\n".join(s)

    def validate(self) -> None:
        """Validate the encoding options and source pixel data."""
        self._validate_options()
        if self.is_dataset or self.is_buffer:
            self._validate_buffer()
        else:
            self._validate_array()

        # UID specific validation based on Section 8 of Part 5
        self._validate_encoding_profile()

    def _validate_array(self) -> None:
        """Check that the ndarray matches the supplied options."""
        arr = cast(np.ndarray, self.src)
        shape = arr.shape
        dtype = arr.dtype

        if len(shape) not in (2, 3, 4):
            raise ValueError(f"Unable to encode {len(shape)}D ndarrays")

        # `arr` may be (for planar configuration 0):
        #   (rows, columns)
        #   (rows, columns, planes)
        #   (frames, rows, columns)
        #   (frames, rows, columns, planes)
        expected = [
            self.number_of_frames,
            self.rows,
            self.columns,
            self.samples_per_pixel,
        ]
        expected = expected[1:] if expected[0] == 1 else expected
        expected = expected[:-1] if expected[-1] in (None, 1) else expected

        if shape != tuple(expected):
            raise ValueError(
                f"Mismatch between the expected ndarray shape {tuple(expected)} "
                f"and the actual shape {shape}"
            )

        # Check dtype is int or uint
        ui = [
            np.issubdtype(dtype, np.unsignedinteger),
            np.issubdtype(dtype, np.signedinteger),
        ]
        if not any(ui):
            raise ValueError(
                f"The ndarray's dtype '{dtype}' is not supported, must be a "
                "signed or unsigned integer type"
            )

        # Check dtype is consistent with the *Pixel Representation*
        is_signed = self.pixel_representation == 1
        if not ui[is_signed]:
            s = ["unsigned", "signed"][is_signed]
            raise ValueError(
                f"The ndarray's dtype '{dtype}' is not consistent with a (0028,0103) "
                f"'Pixel Representation' of '{self.pixel_representation}' ({s} integers)"
            )

        # Check the dtype's itemsize is at least as large as *Bits Allocated*
        if dtype.itemsize < math.ceil(self.bits_allocated / 8):
            raise ValueError(
                f"The ndarray's dtype '{dtype}' is not consistent with a (0028,0100) "
                f"'Bits Allocated' value of '{self.bits_allocated}'"
            )

        # Check the pixel data values are consistent with *Bits Stored*
        amax, amin = arr.max(), arr.min()
        if is_signed:
            minimum = -(2 ** (self.bits_stored - 1))
            maximum = 2 ** (self.bits_stored - 1) - 1
        else:
            minimum, maximum = 0, 2**self.bits_stored - 1

        if amax > maximum or amin < minimum:
            raise ValueError(
                "The ndarray contains values that are outside the allowable "
                f"range of ({minimum}, {maximum}) for a (0028,0101) 'Bits "
                f"Stored' value of '{self.bits_stored}'"
            )

        if self.transfer_syntax == JPEGLSNearLossless and is_signed:
            # JPEG-LS doesn't track signedness, so with lossy encoding of
            #   signed data it's possible to flip from a negative to a positive
            #   value (or vice versa) due to the introduced error.
            # The only way to avoid this is to limit pixel values to the
            #   range (minimum + jls_error, maximum - jls_error), where
            #   `jls_error` is the JPEG-LS 'NEAR' parameter and `minimum`
            #   and `maximum` are the min/max possible values for a given
            #   sample precision
            error = self.get_option("jls_error", 0)
            within = amax <= (maximum - error) and amin >= (minimum + error)
            if error and not within:
                raise ValueError(
                    "The supported range of pixel values when performing lossy "
                    "JPEG-LS encoding of signed integers with a (0028,0103) 'Bits "
                    f"Stored' value of '{self.bits_stored}' and a 'jls_error' "
                    f"of '{error}' is ({minimum + error}, {maximum - error})"
                )

    def _validate_buffer(self) -> None:
        """Validate the supplied pixel data buffer."""
        # Check the length is at least as long as required
        length_bytes = self.frame_length(unit="bytes")
        expected = length_bytes * self.number_of_frames
        if (actual := len(self.src)) < expected:
            raise ValueError(
                "The length of the uncompressed pixel data doesn't match the "
                f"expected length - {actual} bytes actual vs. {expected} expected"
            )

    def _validate_encoding_profile(self) -> None:
        """Perform  UID specific validation of encoding parameters based on
        Part 5, Section 8 of the DICOM Standard.

        Encoding profiles should be:

        Tuple[str, int, Iterable[int], Iterable[int], Iterable[int]] as
        (
            PhotometricInterpretation, SamplesPerPixel, PixelRepresentation,
            BitsAllocated, BitsStored
        )
        """
        if self.transfer_syntax not in ENCODING_PROFILES:
            return

        # Test each profile and see if it matches source parameters
        profile = ENCODING_PROFILES[self.transfer_syntax]
        for pi, spp, px_repr, bits_a, bits_s in profile:
            try:
                assert self.photometric_interpretation == pi
                assert self.samples_per_pixel == spp
                assert self.pixel_representation in px_repr
                assert self.bits_allocated in bits_a
                assert self.bits_stored in bits_s
            except AssertionError:
                continue

            return

        raise ValueError(
            "One or more of the following values is not valid for pixel data "
            f"encoded with '{self.transfer_syntax.name}':\n"
            f"  (0028,0002) Samples per Pixel: {self.samples_per_pixel}\n"
            "  (0028,0006) Photometric Interpretation: "
            f"{self.photometric_interpretation}\n"
            f"  (0028,0100) Bits Allocated: {self.bits_allocated}\n"
            f"  (0028,0101) Bits Stored: {self.bits_stored}\n"
            f"  (0028,0103) Pixel Representation: {self.pixel_representation}\n"
            "See Part 5, Section 8.2 of the DICOM Standard for more information"
        )


class Encoder(CoderBase):
    """Factory class for data encoders.

    Every available ``Encoder`` instance in *pydicom* corresponds directly
    to a single DICOM *Transfer Syntax UID*, and provides a  mechanism for
    converting raw unencoded source data to meet the requirements of that
    transfer syntax using one or more :doc:`encoding plugins
    </guides/encoding/encoder_plugins>`.

    .. versionadded:: 2.2
    """

    def __init__(self, uid: UID) -> None:
        """Create a new data encoder.

        Parameters
        ----------
        uid : pydicom.uid.UID
            The *Transfer Syntax UID* that the encoder supports.
        """
        super().__init__(uid, decoder=False)

    def encode(
        self,
        src: "bytes | np.ndarray | Dataset",
        *,
        index: int | None = None,
        validate: bool = True,
        encoding_plugin: str = "",
        **kwargs: Any,
    ) -> bytes:
        """Return an encoded frame of the pixel data in `src` as
        :class:`bytes`.

        .. warning::

            With the exception of *RLE Lossless*, this method requires the
            installation of additional packages to perform the actual pixel
            data encoding. See the :doc:`encoding documentation
            </guides/user/image_data_compression>` for more information.

        Parameters
        ----------
        src : bytes, numpy.ndarray or pydicom.dataset.Dataset
            Single or multi-frame pixel data as one of the following:

            * :class:`~numpy.ndarray`: the uncompressed pixel data, should be
              :attr:`shaped<numpy.ndarray.shape>` as:

              * (rows, columns) for single frame, single sample data.
              * (rows, columns, planes) for single frame, multi-sample data.
              * (frames, rows, columns) for multi-frame, single sample data.
              * (frames, rows, columns, planes) for multi-frame and
                multi-sample data.

            * :class:`~pydicom.dataset.Dataset`: the dataset containing
              the uncompressed *Pixel Data* to be encoded.
            * :class:`bytes`: the uncompressed little-endian ordered pixel
              data. `src` should use 1, 2, 4 or 8 bytes per pixel, whichever
              of these is sufficient for the (0028,0103) *Bits Stored* value.
        index : int, optional
            Required when `src` contains multiple frames, this is the index
            of the frame to be encoded.
        validate : bool, optional
            If ``True`` (default) then validate the supplied encoding options
            and pixel data prior to encoding, otherwise if ``False`` no
            validation will be performed.
        encoding_plugin : str, optional
            The name of the pixel data encoding plugin to use. If
            `encoding_plugin` is not specified then all available
            plugins will be tried (default). For information on the available
            plugins for each encoder see the
            :mod:`API documentation<pydicom.pixels.encoders>`.
        **kwargs
            The following keyword parameters are required when `src` is
            :class:`bytes` or :class:`~numpy.ndarray`:

            * ``'rows'``: :class:`int` - the number of rows of pixels in `src`,
              maximum 65535.
            * ``'columns'``: :class:`int` - the number of columns of pixels in
              `src`, maximum 65535.
            * ``'number_of_frames'``: :class:`int` - the number of frames
              in `src`.
            * ``'samples_per_pixel'``: :class:`int` - the number of samples
              per pixel in `src`, should be 1 or 3.
            * ``'bits_allocated'``: :class:`int` - the number of bits used
              to contain each pixel, should be a multiple of 8.
            * ``'bits_stored'``: :class:`int` - the number of bits actually
              used per pixel. For example, an ``ndarray`` `src` might have a
              :class:`~numpy.dtype` of ``'uint16'`` (range 0 to 65535) but
              only contain 12-bit pixel values (range 0 to 4095).
            * ``'pixel_representation'``: :class:`int` - the type of data
              being encoded, ``0`` for unsigned, ``1`` for 2's complement
              (signed)
            * ``'photometric_interpretation'``: :class:`str` - the intended
              color space of the *encoded* pixel data, such as ``'YBR_FULL'``.

            Optional keyword parameters for the encoding plugin may also be
            present. See the :doc:`encoding plugin options
            </guides/encoding/encoder_plugin_options>` for more information.

        Returns
        -------
        bytes
            The encoded pixel data.
        """
        if index is not None and index < 0:
            raise ValueError("'index' must be greater than or equal to 0")

        runner = EncodeRunner(self.UID)
        runner.set_source(src)
        runner.set_options(**kwargs)
        runner.set_encoders(
            cast(
                dict[str, "EncodeFunction"],
                self._validate_plugins(encoding_plugin),
            ),
        )

        if config.debugging:
            LOGGER.debug(runner)

        if validate:
            runner.validate()

        if runner.number_of_frames > 1 and index is None:
            raise ValueError(
                "The 'index' of the frame to be encoded is required for "
                "multi-frame pixel data"
            )

        return runner.encode(index)

    def iter_encode(
        self,
        src: "bytes | np.ndarray | Dataset",
        *,
        validate: bool = True,
        encoding_plugin: str = "",
        **kwargs: Any,
    ) -> Iterator[bytes]:
        """Yield encoded frames of the pixel data in `src` as :class:`bytes`.

        .. warning::

            With the exception of *RLE Lossless*, this method requires the
            installation of additional packages to perform the actual pixel
            data encoding. See the :doc:`encoding documentation
            </guides/user/image_data_compression>` for more information.

        Parameters
        ----------
        src : bytes, numpy.ndarray or pydicom.dataset.Dataset
            Single or multi-frame pixel data as one of the following:

            * :class:`~numpy.ndarray`: the uncompressed pixel data, should be
              :attr:`shaped<numpy.ndarray.shape>` as:

              * (rows, columns) for single frame, single sample data.
              * (rows, columns, planes) for single frame, multi-sample data.
              * (frames, rows, columns) for multi-frame, single sample data.
              * (frames, rows, columns, planes) for multi-frame and
                multi-sample data.

            * :class:`~pydicom.dataset.Dataset`: the dataset containing
              the uncompressed *Pixel Data* to be encoded.
            * :class:`bytes`: the uncompressed little-endian ordered pixel
              data. `src` should use 1, 2, 4 or 8 bytes per pixel, whichever
              of these is sufficient for the (0028,0103) *Bits Stored* value.
        validate : bool, optional
            If ``True`` (default) then validate the supplied encoding options
            and pixel data prior to encoding, otherwise if ``False`` no
            validation will be performed.
        encoding_plugin : str, optional
            The name of the pixel data encoding plugin to use. If
            `encoding_plugin` is not specified then all available
            plugins will be tried (default). For information on the available
            plugins for each encoder see the
            :mod:`API documentation<pydicom.pixels.encoders>`.
        **kwargs
            The following keyword parameters are required when `src` is
            :class:`bytes` or :class:`~numpy.ndarray`:

            * ``'rows'``: :class:`int` - the number of rows of pixels in `src`,
              maximum 65535.
            * ``'columns'``: :class:`int` - the number of columns of pixels in
              `src`, maximum 65535.
            * ``'number_of_frames'``: :class:`int` - the number of frames
              in `src`.
            * ``'samples_per_pixel'``: :class:`int` - the number of samples
              per pixel in `src`, should be 1 or 3.
            * ``'bits_allocated'``: :class:`int` - the number of bits used
              to contain each pixel, should be a multiple of 8.
            * ``'bits_stored'``: :class:`int` - the number of bits actually
              used per pixel. For example, an ``ndarray`` `src` might have a
              :class:`~numpy.dtype` of ``'uint16'`` (range 0 to 65535) but
              only contain 12-bit pixel values (range 0 to 4095).
            * ``'pixel_representation'``: :class:`int` - the type of data
              being encoded, ``0`` for unsigned, ``1`` for 2's complement
              (signed)
            * ``'photometric_interpretation'``: :class:`str` - the intended
              color space of the encoded pixel data, such as ``'YBR_FULL'``.

            Optional keyword parameters for the encoding plugin may also be
            present. See the :doc:`encoding plugin options
            </guides/encoding/encoder_plugin_options>` for more information.

        Yields
        ------
        bytes
            An encoded frame of pixel data.
        """
        runner = EncodeRunner(self.UID)
        runner.set_source(src)
        runner.set_options(**kwargs)
        runner.set_encoders(
            cast(
                dict[str, "EncodeFunction"],
                self._validate_plugins(encoding_plugin),
            ),
        )

        if config.debugging:
            LOGGER.debug(runner)

        if validate:
            runner.validate()

        if runner.number_of_frames == 1:
            yield runner.encode(None)
            return

        for index in range(runner.number_of_frames):
            yield runner.encode(index)


# UID: [
#   Photometric Interpretation (the intended value *after* encoding),
#   Samples per Pixel,
#   Pixel Representation,
#   Bits Allocated,
#   Bits Stored,
# ]
ProfileType = tuple[str, int, Iterable[int], Iterable[int], Iterable[int]]
ENCODING_PROFILES: dict[UID, list[ProfileType]] = {
    JPEGBaseline8Bit: [  # 1.2.840.10008.1.2.4.50: Table 8.2.1-1 in PS3.5
        ("MONOCHROME1", 1, (0,), (8,), (8,)),
        ("MONOCHROME2", 1, (0,), (8,), (8,)),
        ("YBR_FULL_422", 3, (0,), (8,), (8,)),
        ("RGB", 3, (0,), (8,), (8,)),
    ],
    JPEGExtended12Bit: [  # 1.2.840.10008.1.2.4.51: Table 8.2.1-1 in PS3.5
        ("MONOCHROME1", 1, (0,), (8,), (8,)),
        ("MONOCHROME1", 1, (0,), (16,), (12,)),
        ("MONOCHROME2", 1, (0,), (8,), (8,)),
        ("MONOCHROME2", 1, (0,), (16,), (12,)),
    ],
    JPEGLossless: [  # 1.2.840.10008.1.2.4.57: Table 8.2.1-2 in PS3.5
        ("MONOCHROME1", 1, (0, 1), (8, 16), range(1, 17)),
        ("MONOCHROME2", 1, (0, 1), (8, 16), range(1, 17)),
        ("PALETTE COLOR", 1, (0,), (8, 16), range(1, 17)),
        ("YBR_FULL", 3, (0,), (8, 16), range(1, 17)),
        ("RGB", 3, (0,), (8, 16), range(1, 17)),
    ],
    JPEGLosslessSV1: [  # 1.2.840.10008.1.2.4.70: Table 8.2.1-2 in PS3.5
        ("MONOCHROME1", 1, (0, 1), (8, 16), range(1, 17)),
        ("MONOCHROME2", 1, (0, 1), (8, 16), range(1, 17)),
        ("PALETTE COLOR", 1, (0,), (8, 16), range(1, 17)),
        ("YBR_FULL", 3, (0,), (8, 16), range(1, 17)),
        ("RGB", 3, (0,), (8, 16), range(1, 17)),
    ],
    JPEGLSLossless: [  # 1.2.840.10008.1.2.4.80: Table 8.2.3-1 in PS3.5
        ("MONOCHROME1", 1, (0, 1), (8, 16), range(2, 17)),
        ("MONOCHROME2", 1, (0, 1), (8, 16), range(2, 17)),
        ("PALETTE COLOR", 1, (0,), (8, 16), range(2, 17)),
        ("YBR_FULL", 3, (0,), (8,), range(2, 9)),
        ("RGB", 3, (0,), (8, 16), range(2, 17)),
    ],
    JPEGLSNearLossless: [  # 1.2.840.10008.1.2.4.81: Table 8.2.3-1 in PS3.5
        ("MONOCHROME1", 1, (0, 1), (8, 16), range(2, 17)),
        ("MONOCHROME2", 1, (0, 1), (8, 16), range(2, 17)),
        ("YBR_FULL", 3, (0,), (8,), range(2, 9)),
        ("RGB", 3, (0,), (8, 16), range(2, 17)),
    ],
    JPEG2000Lossless: [  # 1.2.840.10008.1.2.4.90: Table 8.2.4-1 in PS3.5
        ("MONOCHROME1", 1, (0, 1), (8, 16, 24, 32, 40), range(1, 39)),
        ("MONOCHROME2", 1, (0, 1), (8, 16, 24, 32, 40), range(1, 39)),
        ("PALETTE COLOR", 1, (0,), (8, 16), range(1, 17)),
        ("YBR_RCT", 3, (0,), (8, 16, 24, 32, 40), range(1, 39)),
        ("RGB", 3, (0,), (8, 16, 24, 32, 40), range(1, 39)),
        ("YBR_FULL", 3, (0,), (8, 16, 24, 32, 40), range(1, 39)),
    ],
    JPEG2000: [  # 1.2.840.10008.1.2.4.91: Table 8.2.4-1 in PS3.5
        ("MONOCHROME1", 1, (0, 1), (8, 16, 24, 32, 40), range(1, 39)),
        ("MONOCHROME2", 1, (0, 1), (8, 16, 24, 32, 40), range(1, 39)),
        ("YBR_ICT", 3, (0,), (8, 16, 24, 32, 40), range(1, 39)),
        ("RGB", 3, (0,), (8, 16, 24, 32, 40), range(1, 39)),
        ("YBR_FULL", 3, (0,), (8, 16, 24, 32, 40), range(1, 39)),
    ],
    RLELossless: [  # 1.2.840.10008.1.2.5: Table 8.2.2-1 in PS3.5
        ("MONOCHROME1", 1, (0, 1), (8, 16), range(1, 17)),
        ("MONOCHROME2", 1, (0, 1), (8, 16), range(1, 17)),
        ("PALETTE COLOR", 1, (0,), (8, 16), range(1, 17)),
        ("YBR_FULL", 3, (0,), (8,), range(1, 9)),
        ("RGB", 3, (0,), (8, 16), range(1, 17)),
    ],
}

# Encoder names should be f"{UID.keyword}Encoder"
RLELosslessEncoder = Encoder(RLELossless)
RLELosslessEncoder.add_plugins(
    [
        ("gdcm", ("pydicom.pixels.encoders.gdcm", "encode_pixel_data")),
        ("pylibjpeg", ("pydicom.pixels.encoders.pylibjpeg", "_encode_frame")),
        ("pydicom", ("pydicom.pixels.encoders.native", "_encode_frame")),
    ],
)

JPEGLSLosslessEncoder = Encoder(JPEGLSLossless)
JPEGLSLosslessEncoder.add_plugin(
    "pyjpegls", ("pydicom.pixels.encoders.pyjpegls", "_encode_frame")
)

JPEGLSNearLosslessEncoder = Encoder(JPEGLSNearLossless)
JPEGLSNearLosslessEncoder.add_plugin(
    "pyjpegls", ("pydicom.pixels.encoders.pyjpegls", "_encode_frame")
)

JPEG2000LosslessEncoder = Encoder(JPEG2000Lossless)
JPEG2000LosslessEncoder.add_plugin(
    "pylibjpeg", ("pydicom.pixels.encoders.pylibjpeg", "_encode_frame")
)

JPEG2000Encoder = Encoder(JPEG2000)
JPEG2000Encoder.add_plugin(
    "pylibjpeg", ("pydicom.pixels.encoders.pylibjpeg", "_encode_frame")
)


# Available pixel data encoders
_PIXEL_DATA_ENCODERS = {
    # UID: (encoder, 'versionadded')
    RLELossless: (RLELosslessEncoder, "2.2"),
    JPEGLSLossless: (JPEGLSLosslessEncoder, "3.0"),
    JPEGLSNearLossless: (JPEGLSNearLosslessEncoder, "3.0"),
    JPEG2000Lossless: (JPEG2000LosslessEncoder, "3.0"),
    JPEG2000: (JPEG2000Encoder, "3.0"),
}


def _build_encoder_docstrings() -> None:
    """Override the default Encoder docstring."""
    plugin_doc_links = {
        "pydicom": ":ref:`pydicom <encoder_plugin_pydicom>`",
        "pylibjpeg": ":ref:`pylibjpeg <encoder_plugin_pylibjpeg>`",
        "gdcm": ":ref:`gdcm <encoder_plugin_gdcm>`",
        "pyjpegls": ":ref:`pyjpegls <encoder_plugin_pyjpegls>`",
    }

    for enc, versionadded in _PIXEL_DATA_ENCODERS.values():
        uid = enc.UID
        available = enc._available.keys()
        unavailable = enc._unavailable.keys()
        plugins = list(available) + list(unavailable)

        plugins = [plugin_doc_links[name] for name in sorted(plugins)]

        s = [f"A *Pixel Data* encoder for *{uid.name}* - ``{uid}``"]
        s.append("")
        s.append(f".. versionadded:: {versionadded}")
        s.append("")
        s.append(f"Encoding plugins: {', '.join(plugins)}")
        s.append("")
        s.append(
            "See the :class:`~pydicom.pixels.encoders.base.Encoder` "
            "reference for instance methods and attributes."
        )
        enc.__doc__ = "\n".join(s)


_build_encoder_docstrings()


def get_encoder(uid: str) -> Encoder:
    """Return the pixel data encoder corresponding to `uid`.

    .. versionadded:: 2.2

    +--------------------------------------------------+----------------+
    | Transfer Syntax                                  | Version added  |
    +-------------------------+------------------------+                +
    | Name                    | UID                    |                |
    +=========================+========================+================+
    | *JPEG-LS Lossless*      | 1.2.840.10008.1.2.4.80 | 3.0            |
    +-------------------------+------------------------+----------------+
    | *JPEG-LS Near Lossless* | 1.2.840.10008.1.2.4.81 | 3.0            |
    +-------------------------+------------------------+----------------+
    | *JPEG 2000 Lossless*    | 1.2.840.10008.1.2.4.90 | 3.0            |
    +-------------------------+------------------------+----------------+
    | *JPEG 2000*             | 1.2.840.10008.1.2.4.91 | 3.0            |
    +-------------------------+------------------------+----------------+
    | *RLE Lossless*          | 1.2.840.10008.1.2.5    | 2.2            |
    +-------------------------+------------------------+----------------+
    """
    uid = UID(uid)
    try:
        return _PIXEL_DATA_ENCODERS[uid][0]
    except KeyError:
        raise NotImplementedError(
            f"No pixel data encoders have been implemented for '{uid.name}'"
        )



# Copyright 2008-2024 pydicom authors. See LICENSE file for details.

from pydicom.pixels.encoders.base import (
    RLELosslessEncoder,
    JPEGLSLosslessEncoder,
    JPEGLSNearLosslessEncoder,
    JPEG2000LosslessEncoder,
    JPEG2000Encoder,
)



# Copyright 2020 pydicom authors. See LICENSE file for details.
"""Pydicom command line interface program

Each subcommand is a module within pydicom.cli, which
defines an add_subparser(subparsers) function to set argparse
attributes, and calls set_defaults(func=callback_function)

"""

import argparse
from importlib.metadata import entry_points
import re
import sys
from typing import cast, Any
from collections.abc import Callable

from pydicom import dcmread
from pydicom.data.data_manager import get_charset_files, get_testdata_file
from pydicom.dataset import Dataset


subparsers: argparse._SubParsersAction | None = None


# Restrict the allowed syntax tightly, since use Python `eval`
# on the expression. Do not allow callables, or assignment, for example.
re_kywd_or_item = (
    r"\w+"  # Keyword (\w allows underscore, needed for file_meta)
    r"(\[(-)?\d+\])?"  # Optional [index] or [-index]
)

re_file_spec_object = re.compile(re_kywd_or_item + r"(\." + re_kywd_or_item + r")*$")

filespec_help = (
    "File specification, in format [pydicom::]filename[::element]. "
    "If `pydicom::` prefix is present, then use the pydicom "
    "test file with that name. If `element` is given, "
    "use only that data element within the file. "
    "Examples: "
    "path/to/your_file.dcm, "
    "your_file.dcm::StudyDate, "
    "pydicom::rtplan.dcm::BeamSequence[0], "
    "yourplan.dcm::BeamSequence[0].BeamNumber"
)


def eval_element(ds: Dataset, element: str) -> Any:
    try:
        return eval("ds." + element, {"ds": ds})
    except AttributeError:
        raise argparse.ArgumentTypeError(
            f"Data element '{element}' is not in the dataset"
        )
    except IndexError as e:
        raise argparse.ArgumentTypeError(f"'{element}' has an index error: {e}")


def filespec_parts(filespec: str) -> tuple[str, str, str]:
    """Parse the filespec format into prefix, filename, element

    Format is [prefix::filename::element]

    Note that ':' can also exist in valid filename, e.g. r'c:\temp\test.dcm'
    """

    *prefix_file, last = filespec.split("::")

    if not prefix_file:  # then only the filename component
        return "", last, ""

    prefix = "pydicom" if prefix_file[0] == "pydicom" else ""
    if prefix:
        prefix_file.pop(0)

    # If list empty after pop above, then have pydicom::filename
    if not prefix_file:
        return prefix, last, ""

    return prefix, "".join(prefix_file), last


def filespec_parser(filespec: str) -> list[tuple[Dataset, Any]]:
    """Utility to return a dataset and an optional data element value within it

    Note: this is used as an argparse 'type' for adding parsing arguments.

    Parameters
    ----------
    filespec: str
        A filename with optional `pydicom::` prefix and optional data element,
        in format:
            [pydicom::]<filename>[::<element>]
        If an element is specified, it must be a path to a data element,
        sequence item (dataset), or a sequence.
        Examples:
            your_file.dcm
            your_file.dcm::StudyDate
            pydicom::rtplan.dcm::BeamSequence[0]
            pydicom::rtplan.dcm::BeamSequence[0].BeamLimitingDeviceSequence

    Returns
    -------
    List[Tuple[Dataset, Any]]
        Matching pairs of (dataset, data element value)
        This usually is a single pair, but a list is returned for future
        ability to work across multiple files.

    Note
    ----
        This function is meant to be used in a call to an `argparse` library's
        `add_argument` call for subparsers, with name="filespec" and
        `type=filespec_parser`. When used that way, the resulting args.filespec
        will contain the return values of this function
        (e.g. use `ds, element_val = args.filespec` after parsing arguments)
        See the `pydicom.cli.show` module for an example.

    Raises
    ------
    argparse.ArgumentTypeError
        If the filename does not exist in local path or in pydicom test files,
        or if the optional element is not a valid expression,
        or if the optional element is a valid expression but does not exist
        within the dataset
    """
    prefix, filename, element = filespec_parts(filespec)

    # Get the pydicom test filename even without prefix, in case user forgot it
    try:
        pydicom_filename = cast(str, get_testdata_file(filename))
    except ValueError:  # will get this if absolute path passed
        pydicom_filename = ""

    # Check if filename is in charset files
    if not pydicom_filename:
        try:
            char_filenames = get_charset_files(filename)
            if char_filenames:
                pydicom_filename = char_filenames[0]
        except NotImplementedError:  # will get this if absolute path passed
            pass

    if prefix == "pydicom":
        filename = pydicom_filename

    # Check element syntax first to avoid unnecessary load of file
    if element and not re_file_spec_object.match(element):
        raise argparse.ArgumentTypeError(
            f"Component '{element}' is not valid syntax for a "
            "data element, sequence, or sequence item"
        )

    # Read DICOM file
    try:
        ds = dcmread(filename, force=True)
    except FileNotFoundError:
        extra = (
            (f", \nbut 'pydicom::{filename}' test data file is available")
            if pydicom_filename
            else ""
        )
        raise argparse.ArgumentTypeError(f"File '{filename}' not found{extra}")
    except Exception as e:
        raise argparse.ArgumentTypeError(f"Error reading '{filename}': {e}")

    if not element:
        return [(ds, None)]

    data_elem_val = eval_element(ds, element)

    return [(ds, data_elem_val)]


def help_command(args: argparse.Namespace) -> None:
    if subparsers is None:
        print("No subcommands are available")
        return

    subcommands: list[str] = list(subparsers.choices.keys())
    if args.subcommand and args.subcommand in subcommands:
        subparsers.choices[args.subcommand].print_help()
    else:
        print("Use pydicom help [subcommand] to show help for a subcommand")
        subcommands.remove("help")
        print(f"Available subcommands: {', '.join(subcommands)}")


SubCommandType = dict[str, Callable[[argparse._SubParsersAction], None]]


def get_subcommand_entry_points() -> SubCommandType:
    subcommands = {}
    for entry_point in entry_points(group="pydicom_subcommands"):
        subcommands[entry_point.name] = entry_point.load()

    return subcommands


def main(args: list[str] | None = None) -> None:
    """Entry point for 'pydicom' command line interface

    Parameters
    ----------
    args : List[str], optional
        Command-line arguments to parse.  If ``None``, then :attr:`sys.argv`
        is used.
    """
    global subparsers

    py_version = sys.version.split()[0]

    parser = argparse.ArgumentParser(
        prog="pydicom",
        description=f"pydicom command line utilities (Python {py_version})",
    )
    subparsers = parser.add_subparsers(help="subcommand help")

    help_parser = subparsers.add_parser("help", help="display help for subcommands")
    help_parser.add_argument(
        "subcommand", nargs="?", help="Subcommand to show help for"
    )
    help_parser.set_defaults(func=help_command)

    # Get subcommands to register themselves as a subparser
    subcommands = get_subcommand_entry_points()
    for subcommand in subcommands.values():
        subcommand(subparsers)

    ns = parser.parse_args(args)
    if not vars(ns):
        parser.print_help()
    else:
        ns.func(ns)



# Copyright 2020 pydicom authors. See LICENSE file for details.
"""Pydicom command line interface program for codify"""

import argparse

import pydicom.util.codify


default_exclude_size = 100


def add_subparser(subparsers: argparse._SubParsersAction) -> None:
    codify_parser = subparsers.add_parser(
        "codify",
        description=(
            "Read a DICOM file and produce the pydicom (Python) "
            "code which can create that file"
        ),
        epilog=(
            "Binary data (e.g. pixels) larger than --exclude-size "
            f"(default {default_exclude_size} bytes) is not included. "
            "A dummy line with a syntax error is produced. "
            "Private data elements are not included by default."
        ),
    )

    # Codify existed before as a stand-alone before, re-use it here
    pydicom.util.codify.set_parser_arguments(codify_parser, default_exclude_size)
    codify_parser.set_defaults(func=pydicom.util.codify.do_codify)



# Copyright 2008-2024 pydicom authors. See LICENSE file for details.
"""Use Pillow <https://github.com/python-pillow/Pillow> to decompress encoded
*Pixel Data*.

This module is not intended to be used directly.
"""

from io import BytesIO
from typing import cast

from pydicom import uid
from pydicom.pixels.utils import _passes_version_check
from pydicom.pixels.common import PhotometricInterpretation as PI
from pydicom.pixels.decoders.base import DecodeRunner

try:
    from PIL import Image, features
except ImportError:
    pass

try:
    import numpy as np

    HAVE_NP = True
except ImportError:
    HAVE_NP = False


DECODER_DEPENDENCIES = {
    uid.JPEGBaseline8Bit: ("pillow>=10.0",),
    uid.JPEGExtended12Bit: ("pillow>=10.0",),
    uid.JPEG2000Lossless: ("numpy", "pillow>=10.0"),
    uid.JPEG2000: ("numpy", "pillow>=10.0"),
}

_LIBJPEG_SYNTAXES = [uid.JPEGBaseline8Bit, uid.JPEGExtended12Bit]
_OPENJPEG_SYNTAXES = [uid.JPEG2000Lossless, uid.JPEG2000]


def is_available(uid: str) -> bool:
    """Return ``True`` if a pixel data decoder for `uid` is available for use,
    ``False`` otherwise.
    """
    if not _passes_version_check("PIL", (10, 0)):
        return False

    if uid in _LIBJPEG_SYNTAXES:
        return bool(features.check_codec("jpg"))  # type: ignore[no-untyped-call]

    if uid in _OPENJPEG_SYNTAXES:
        return bool(features.check_codec("jpg_2000")) and HAVE_NP  # type: ignore[no-untyped-call]

    return False


def _decode_frame(src: bytes, runner: DecodeRunner) -> bytes:
    """Return the decoded image data in `src` as a :class:`bytes`."""
    tsyntax = runner.transfer_syntax

    # libjpeg only supports 8-bit JPEG Extended (can be 8 or 12 in the JPEG standard)
    if tsyntax == uid.JPEGExtended12Bit and runner.bits_stored != 8:
        raise NotImplementedError(
            "Pillow does not support 'JPEG Extended' for samples with 12-bit precision"
        )

    image = Image.open(BytesIO(src), formats=("JPEG", "JPEG2000"))
    if tsyntax in _LIBJPEG_SYNTAXES:
        if runner.samples_per_pixel != 1:
            # If the Adobe APP14 marker is not present then Pillow assumes
            #   that JPEG images were transformed into YCbCr color space prior
            #   to compression, so setting the image mode to YCbCr signals we
            #   don't want any color transformations.
            # Any color transformations would be inconsistent with the
            #   behavior required by the `raw` flag
            if "adobe_transform" not in image.info:
                image.draft("YCbCr", image.size)  # type: ignore[no-untyped-call]

        return cast(bytes, image.tobytes())

    # JPEG 2000
    # The precision from the J2K codestream is more appropriate because the
    #   decoder will use it to create the output integers
    precision = runner.get_option("j2k_precision", runner.bits_stored)
    # pillow's pixel container size is based on precision
    if 0 < precision <= 8:
        runner.set_option("bits_allocated", 8)
    elif 8 < precision <= 16:
        # Pillow converts >= 9-bit RGB/YCbCr data to 8-bit
        if runner.samples_per_pixel > 1:
            raise ValueError(
                f"Pillow cannot decode {precision}-bit multi-sample data correctly"
            )

        runner.set_option("bits_allocated", 16)
    else:
        raise ValueError(
            "only (0028,0101) 'Bits Stored' values of up to 16 are supported"
        )

    # Pillow converts N-bit signed/unsigned data to 8- or 16-bit unsigned data
    #   See Pillow src/libImaging/Jpeg2KDecode.c::j2ku_gray_i
    buffer = bytearray(image.tobytes())  # so the array is writeable
    del image
    dtype = runner.pixel_dtype
    arr = np.frombuffer(buffer, dtype=f"u{dtype.itemsize}")

    is_signed = runner.pixel_representation
    if runner.get_option("apply_j2k_sign_correction", False):
        is_signed = runner.get_option("j2k_is_signed", is_signed)

    if is_signed and runner.pixel_representation == 1:
        # Re-view the unsigned integers as signed
        #   e.g. [0, 127, 128, 255] -> [0, 127, -128, -1]
        arr = arr.view(dtype)
        # Level-shift to match the unsigned integers range
        #   e.g. [0, 127, -128, -1] -> [-128, -1, 0, 127]
        arr -= np.int32(2 ** (runner.bits_allocated - 1))

    if bit_shift := (runner.bits_allocated - precision):
        # Bit shift to undo the upscaling of N-bit to 8- or 16-bit
        np.right_shift(arr, bit_shift, out=arr)

    # pillow returns YBR_ICT and YBR_RCT as RGB
    if runner.photometric_interpretation in (PI.YBR_ICT, PI.YBR_RCT):
        runner.set_option("photometric_interpretation", PI.RGB)

    return cast(bytes, arr.tobytes())



# Copyright 2008-2024 pydicom authors. See LICENSE file for details.

from pydicom.pixels.decoders.base import (
    ImplicitVRLittleEndianDecoder,
    ExplicitVRLittleEndianDecoder,
    ExplicitVRBigEndianDecoder,
    DeflatedExplicitVRLittleEndianDecoder,
    JPEGBaseline8BitDecoder,
    JPEGExtended12BitDecoder,
    JPEGLosslessDecoder,
    JPEGLosslessSV1Decoder,
    JPEGLSLosslessDecoder,
    JPEGLSNearLosslessDecoder,
    JPEG2000LosslessDecoder,
    JPEG2000Decoder,
    HTJ2KLosslessDecoder,
    HTJ2KLosslessRPCLDecoder,
    HTJ2KDecoder,
    RLELosslessDecoder,
)



"""Pure python package for DICOM medical file reading and writing."""

import re
from typing import cast
from re import Match
from importlib.metadata import version

__version__: str = version("pydicom")

result = cast(Match[str], re.match(r"(\d+\.\d+\.\d+).*", __version__))
__version_info__ = tuple(result.group(1).split("."))


# DICOM Standard version used for:
#   _dicom_dict.py, _uid_dict.py and uid.py
__dicom_version__: str = "2024c"


# DICOM Standard version used for:
#   sr/_cid_dict.py
#   sr/_concepts_dict.py
#   sr/_snomed_dict.py
__concepts_version__: str = "2024c"
